{*****************************************************************************
* UNIT DESCRIPTION:  Form class for charting
* SPEC REF: (if used)
* DESIGN REF: (if used)
*
* Copyright (c) 2009 Schweitzer Engineering Laboratories, Pullman, Washington
******************************************************************************}
unit FormChartTimeSeries;

interface

uses
  Windows, Messages, SysUtils, StrUtils, DateUtils,
  Controls, Tabs, Forms, Dialogs, Buttons, ExtCtrls, Graphics,
  Classes, Contnrs, ToolWin, ComCtrls, StdCtrls, ClipBrd,
  TeeEdit, TeEngine, TeeTools, TeeProcs, Chart, Series,
  ClassTimeSeries, FormChartDeviceInfo, ClassTimeSeriesChart,
  ClassTimeSeriesChartSSI, ClassTimeSeriesChartLDP, ClassChartInput,
  ClassChartInputSSI, ClassChartInputLDP, TimeRoutines, Types, FormChartNotepad;
const
  INVALID_CHART_MSG = 'Invalid chart type.';
  CRLF = Chr(13) + Chr(10);     // carriage return line feed
  CR = Chr(13);     // carriage return
  TAB = Chr(9);
  INTERVAL_COLOR = clWebLightSlateGray;
  PARENTEVENT_COLOR = clYellow;
  SUBEVENT_COLOR = clRed;
  EVENTMARKER_COLOR = clYellow;
  DATETIME_WITH_MS = 'MM/dd/yy hh:mm:ss.zzz';
  DATETIME_WITH_SEC = 'MM/dd/yy hh:mm:ss';
  TIME_WITH_MS = 'hh:mm:ss.zzz';
  TIME_WITH_SEC = 'hh:mm:ss';
  SHGFP_TYPE_CURRENT         = 0;
  CSIDL_PERSONAL = $0005;  // Doc and settings > User > My Documents
type
  TChartMenuEvent = procedure (form : TForm) of object;  // method pointer

  TSystemFolderFunction = function (hwnd: HWND; csidl: Integer; hToken: THandle;
          dwFlags: DWORD; pszPath: PChar): HResult; stdcall;

  ESystemFolderException = class(Exception);

  TfrmChartTimeSeries = class(TForm)
    PanelEventList: TPanel;
    PanelChart: TPanel;
    SplitterEventForm: TSplitter;
    ScrollBoxEventSelection: TScrollBox;
    ListViewEventList: TListView;
    LabelStatus: TLabel;
    GroupBoxEventDateTime: TGroupBox;
    LabelEventType: TLabel;
    LabelEventPhase: TLabel;
    CheckBoxEventBegin: TCheckBox;
    CheckBoxEventEnd: TCheckBox;
    DateTimePickerEventBeginDate: TDateTimePicker;
    ComboBoxEventPhase: TComboBox;
    ButtonEventUpdate: TButton;
    DateTimePickerEventEndDate: TDateTimePicker;
    DateTimePickerEventEndTime: TDateTimePicker;
    DateTimePickerEventBeginTime: TDateTimePicker;
    ComboBoxEventType: TComboBox;
    ScrollBoxChartControls: TScrollBox;
    LabelChartType: TLabel;
    ComboBoxChartType: TComboBox;
    LabelChartTransparency: TLabel;
    ScrollBarChartTransparency: TScrollBar;
    ToolBar1: TToolBar;
    SpeedButtonPointer: TSpeedButton;
    SpeedButtonRotate: TSpeedButton;
    SpeedButtonTranslate: TSpeedButton;
    SpeedButtonZoom: TSpeedButton;
    SpeedButtonThickness: TSpeedButton;
    SpeedButton3D: TSpeedButton;
    SpeedButtonEdit: TSpeedButton;
    SpeedButtonCopy: TSpeedButton;
    SpeedButtonReset: TSpeedButton;
    LabelChartToolbar: TLabel;
    ChartEditorHMIDisplay: TChartEditor;
    SpeedButtonCursor: TSpeedButton;
    SpeedButtonDateTime: TSpeedButton;
    TabSetChart: TTabSet;
    ChartHMIDisplay: TChart;
    ChartTool1: TCursorTool;
    ChartHMICBEMA: TChart;
    ChartTool2: TCursorTool;
    sbUnselectAll: TSpeedButton;
    sbSelectAll: TSpeedButton;
    CheckBoxMinMax: TCheckBox;
    ButtonEventCapture: TSpeedButton;
    SpeedButtonLabel: TSpeedButton;
    SpeedButtonInformation: TSpeedButton;
    LabelScale: TLabel;
    ComboBoxScale: TComboBox;
    LabeDateFormat: TLabel;
    ComboBoxDateFormat: TComboBox;
    ButtonImport: TButton;
    ButtonNotepad: TButton;
    procedure ButtonNotepadClick(Sender: TObject);
    procedure ComboBoxDateFormatChange(Sender: TObject);
    procedure ComboBoxScaleChange(Sender: TObject);
    procedure FormHide(Sender: TObject);
    procedure ChartHMIDisplayClickLegend(Sender: TCustomChart;
      Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure SpeedButtonInformationClick(Sender: TObject);
    procedure SpeedButtonLabelClick(Sender: TObject);
    procedure ListViewEventListSelectItem(Sender: TObject; Item: TListItem;
      Selected: Boolean);
    procedure FormResize(Sender: TObject);
    procedure sbSelectAllClick(Sender: TObject);
    procedure ButtonImportClick(Sender: TObject);
    procedure TabSetChartChange(Sender: TObject; NewTab: Integer;
                                var AllowChange: Boolean);
    procedure ComboBoxEventPhaseExit(Sender: TObject);
    procedure ComboBoxEventTypeExit(Sender: TObject);
    procedure SpeedButtonDateTimeClick(Sender: TObject);
    procedure SpeedButtonCursorClick(Sender: TObject);
    procedure SpeedButtonZoomClick(Sender: TObject);
    procedure SpeedButton3DClick(Sender: TObject);
    procedure SpeedButtonRotateClick(Sender: TObject);
    procedure SpeedButtonResetClick(Sender: TObject);
    procedure SpeedButtonTranslateClick(Sender: TObject);
    procedure SpeedButtonThicknessClick(Sender: TObject);
    procedure SpeedButtonEditClick(Sender: TObject);
    procedure SpeedButtonCopyClick(Sender: TObject);
    procedure SpeedButtonPointerClick(Sender: TObject);
    procedure ScrollBarChartTransparencyChange(Sender: TObject);
    procedure ChartHMIDisplayGetNextAxisLabel(Sender: TChartAxis;
      LabelIndex: Integer; var LabelValue: Double; var Stop: Boolean);
    procedure ButtonEventClearClick(Sender: TObject);
    procedure ListViewEventListCompare(Sender: TObject; Item1, Item2: TListItem;
      Data: Integer; var Compare: Integer);
    procedure ListViewEventListColumnClick(Sender: TObject;
      Column: TListColumn);
    procedure ButtonEventCaptureClick(Sender: TObject);
    procedure ChartHMIDisplayMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure ChartHMIDisplayMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure ChartHMIDisplayMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure ChartHMIDisplayGetAxisLabel(Sender: TChartAxis;
      Series: TChartSeries; ValueIndex: Integer; var LabelText: string);
    procedure FormMouseWheel(Sender: TObject; Shift: TShiftState;
      WheelDelta: Integer; MousePos: TPoint; var Handled: Boolean);
    procedure ChartHMIDisplayMouseWheel(Sender: TObject; Shift: TShiftState;
      WheelDelta: Integer; MousePos: TPoint; var Handled: Boolean);
    procedure ButtonChartResetClick(Sender: TObject);
    procedure ComboBoxChartTypeChange(Sender: TObject);
    procedure ListViewEventListMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
    procedure CheckBoxEventEndClick(Sender: TObject);
    procedure CheckBoxEventBeginClick(Sender: TObject);
    procedure ButtonEventUpdateClick(Sender: TObject);
  private
    FChartType              : TTimeSeriesChartType;   // chart type: ctLDP, ctSSI
    FDeviceType             : TTimeSeriesDeviceType;  // device type: dtSEL7XX...
    FChartTimeSeries        : TTimeSeriesChart;
    FDeviceInfoDialog       : TFormDeviceInfo;     // information window
    ChartSeries             : array of TCustomSeries;
    BarSeries               : TBarSeries;
    PointSeries             : TPointSeries;  // used for LDP events
    CBEMALimitSeries        : array [0..1] of TLineSeries;
//    CBEMAFunctionSeries   : array [0..1] of TLineSeries;  // USE LATER
    CBEMADataSeries         : TPointSeries;
    FChartInputSettingsSSI  : TTimeSeriesInputSettingsSSI;
    FChartInputSettingsLDP  : TTimeSeriesInputSettingsLDP;
    FDateFormat             : string;  // date format string
    FNotepadForm            : TFormNotepad;
    FMenuChart              : boolean;   // true if launched from tools menu
    FChartInstance          : integer;  // index identifying chart
  public
    ChartMenuHandler : TChartMenuEvent; // delegate (method pointer)

    Constructor Create(aOwner: TComponent); reintroduce; overload;
    Destructor Destroy();  overload; override;
    function ReportType(filePath : string) : TTimeSeriesChartType;
    procedure SetChartType(aChartType : TTimeSeriesChartType);
    function  GetChartFilename(aChartType : TTimeSeriesChartType; aFileType : TTimeSeriesFileType; const ExportPath, aID : string) : string;
    function  TestChartData(aChartType : TTimeSeriesChartType; filePath : string; var HasData : boolean) : boolean;
    function  LoadData(filePath : string; var HasData : boolean): boolean;
    function  ReadChartData(aChartType : TTimeSeriesChartType;
                            fileReset: boolean; filePath : string;
                            var ChartReset, ChartValid : boolean ) : boolean;
    procedure ResetInput();
    procedure ResetChartDisplay();
    procedure UpdateChart(reset : boolean);   overload;
    procedure UpdateChart;  overload;
    procedure UpdateActiveEvents;
    function  DisplayChartFile(filePath : string) : boolean;
    function  DisplayChart(report : string) : boolean; overload;
    function  DisplayChart(aChartType : TTimeSeriesChartType; fileReset: boolean; filePath : string; showChart : boolean ) : boolean; overload;
    procedure DisplayChart(found,chartReset,showChart,fileReset,doCBEMA : boolean);  overload;
    procedure FillEventListViewSSI(aReset : boolean);
    procedure FillEventListViewLDP(aReset : boolean);
    procedure Display(reset : boolean);
    procedure DisplayCBEMA(reset : boolean);
    procedure DisplayCBEMALimits(maxDuration : double);
    function  GetActiveEventCount : integer;
    procedure InitChartMinMax(reset : boolean; var xMin,xMax,yMin,yMax : double);
    procedure UpdateChartMinMax(reset : boolean; xMin,xMax,yMin,yMax : double);
    procedure InitChartAttributes(channelCount : integer);
    procedure InitChartAttributesCBEMA;
    procedure DisplaySEREvents(checkCount : integer);
    procedure RecordEventMarkers(iEvent, parentEvent: integer;
                             index,start,stop,estart,estop : integer;
                             myRecord : double;
                             var chartMarker,eventNumber : array of integer;
                             var chartHeight : array of single;
                             var eventColor : array of TColor;
                             var markCount : integer);
    procedure DisplayEventMarkers(checkCount,markCount : integer;
                             chartMarker,eventNumber : array of integer;
                             chartHeight : array of single;
                             eventColor : array of TColor);
    procedure SortEventsByDuration(eventCount : integer;
                               var sDuration, sAmplitude : array of double;
                               var sEvent : array of integer;
                               var sEventType : array of TTimeSeriesCBEMAViolationType);
    function GetCBEMA : boolean;
    procedure SetCBEMA(doCBEMA : boolean);
    procedure SetChartLabels();
    procedure SetChartLineStyles();
    procedure SetDateTimeLabels();
    procedure ResetDeviceInformation;
    procedure ClearChartWindow;
    procedure ResetChartWindow;
    procedure ResetChartControls();
    procedure ShowDeviceInformation;
    procedure ShowChannelInformation;
    procedure ShowEventInformation;
    procedure SetChartDateRangeFilters();
    function CBEMAChartVisible(DoCBEMA : boolean) : boolean;
    procedure WMExitSizeMove(var Message: TMessage) ; message WM_EXITSIZEMOVE;
    procedure SetChartInputControlSettingsSSI();
    procedure GetChartInputControlSettingsSSI();
    procedure SetChartInputControlSettingsLDP();
    procedure GetChartInputControlSettingsLDP();

    function TrimFileName(filename : string) : string;
    function AppendCSVHeader(var MeterBuffer : string) : boolean;
    function AppendCSVData(var MeterBuffer : string) : boolean;
    function AppendReport(var MeterBuffer : string) : boolean;
    function FillEventIndexTable(filter : boolean) : boolean;
    procedure SaveStringToFile(filePath, aString : string);
    procedure RenameOldFile(filePath : string);
    procedure InitChart(aFileType : TTimeSeriesFileType);
    procedure EraseChartDisplay;
    procedure DisplayMouseQueryFeedback;
    procedure SetDateFormat(aDateFormat: string);
    function  GetDateFormat(aDateFormat: string) : string;

    function ChartSeriesActiveState(Num : Integer) : boolean;  // pointer method
    procedure UpdateChartHandler();
    procedure UpdateReportHandler();
    function GetSpecialFolder(csidl: Integer): String;

    property ChartType : TTimeSeriesChartType read FChartType;
    property DeviceType : TTimeSeriesDeviceType read FDeviceType;
    property ChartTimeSeries : TTimeSeriesChart read FChartTimeSeries;
    property DeviceInfoDialog : TFormDeviceInfo read FDeviceInfoDialog write FDeviceInfoDialog;
    property ChartInputSettingsSSI : TTimeSeriesInputSettingsSSI read FChartInputSettingsSSI;
    property ChartInputSettingsLDP : TTimeSeriesInputSettingsLDP read FChartInputSettingsLDP;
    property DateFormat : string read FDateFormat write FDateFormat;
    property MenuChart : boolean read FMenuChart write FMenuChart;
    property ChartInstance : integer read FChartInstance write FChartInstance;

  end;

implementation

Uses EditChar, TeeEditCha, Math;

{$R *.dfm}
{************************************************************************
* NAME:   Create
* FUNCTIONAL DESCRIPTION: class constructor
* INPUT PARAMETERS:  none
* OUTPUT PARAMETERS: none
* RETURN:            none
************************************************************************}
constructor TfrmChartTimeSeries.Create(aOwner: TComponent);
begin
  inherited Create(aOwner);
  FChartTimeSeries := TTimeSeriesChartLDP.Create(nil);
  FChartType := ctInvalid;    // nothing charted yet
  FDeviceInfoDialog := TFormDeviceInfo.Create(nil);
  FDeviceInfoDialog.Parent :=  self;
  FDeviceInfoDialog.Visible := false;
  FDateFormat := 'mm/dd/yyyy';

  self.HelpKeyword := 'Chart';
  dateTimePickerEventBeginDate.DateTime := Now;
  dateTimePickerEventEndDate.DateTime := Now;
  ScrollBarChartTransparency.Enabled := false;
  ChartTool1.Active := false;
  ChartTool1.Snap := true;
  ChartTool2.Active := false;
  ChartTool2.Snap := true;
  FChartInstance := 0;
  FMenuChart := false;
end;
{************************************************************************
* NAME:   Destroy
* FUNCTIONAL DESCRIPTION: class destructor
* INPUT PARAMETERS:  none
* OUTPUT PARAMETERS: none
* RETURN:            none
************************************************************************}
destructor TfrmChartTimeSeries.Destroy();
begin
  ChartHMIDisplay.FreeAllSeries;
  ChartHMICBEMA.FreeAllSeries;
  FreeAndNil(FDeviceInfoDialog);
  FreeAndNil(FChartTimeSeries);

  if(FChartInputSettingsSSI <> nil) then
  begin
    FChartInputSettingsSSI.Free;
  end;
  if(FChartInputSettingsLDP <> nil) then
  begin
    FChartInputSettingsLDP.Free;
  end;

  inherited Destroy();
end;
{************************************************************************
* NAME:  InitChart
* FUNCTIONAL DESCRIPTION: Initialize chart class members before importing report file
* INPUT PARAMETERS: none
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************}
procedure TfrmChartTimeSeries.InitChart(aFileType : TTimeSeriesFileType);
begin
  FChartTimeSeries.InitChart(afileType);
end;
{************************************************************************
* NAME:     EraseChartDisplay
* FUNCTIONAL DESCRIPTION:  Erase content of chart
* INPUT PARAMETERS: none
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************}
procedure TfrmChartTimeSeries.EraseChartDisplay();
begin
 FChartTimeSeries.EraseChartDisplay;
end;
{************************************************************************
* NAME:     ClearChartWindow
* FUNCTIONAL DESCRIPTION:  Erase content of chart and reset window
* INPUT PARAMETERS: none
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************}
procedure TfrmChartTimeSeries.ClearChartWindow();
begin
  FChartTimeSeries.EraseChartDisplay;
  FChartType := ctUndefined;
  ResetChartWindow;
end;
{************************************************************************
* NAME:   TrimFileName
* FUNCTIONAL DESCRIPTION:  trim filename from the end of a file path name
* INPUT PARAMETERS: filename : string  is the original filename
* OUTPUT PARAMETERS: none
* RETURN:             result : string is the trimmed pathname
************************************************************************}
function TfrmChartTimeSeries.TrimFileName(filename : string) : string;
begin
  result := FChartTimeSeries.TrimFileName(filename);
end;
{************************************************************************
* NAME:   SaveStringToFile
* FUNCTIONAL DESCRIPTION: Save string to given file name
* INPUT PARAMETERS:
*                    filePath : string  name of file to save string to
*                    aString : string   string to save
* OUTPUT PARAMETERS: none
* RETURN:            none
************************************************************************}
procedure TfrmChartTimeSeries.SaveStringToFile(filePath, aString : string);
begin
  // Save data in file so that it may be read back as individual records
  FChartTimeSeries.SaveStringToFile(filePath,aString);
end;
{************************************************************************
* NAME:   RenameOldFile
* FUNCTIONAL DESCRIPTION: Rename file if it exists
* INPUT PARAMETERS:
*                    filePath : string  name of file to save string to
* OUTPUT PARAMETERS: none
* RETURN:            none
************************************************************************}
procedure TfrmChartTimeSeries.RenameOldFile(filePath : string);
begin
  // Save data in file so that it may be read back as individual records
  FChartTimeSeries.RenameOldFile(filePath);
end;
{************************************************************************
* NAME:    AppendCSVHeader
* FUNCTIONAL DESCRIPTION:  Append header records for CSV to string
* INPUT PARAMETERS: none
* OUTPUT PARAMETERS: MeterBuffer : string  file content
* RETURN: result : boolean   status of export (true = success)
************************************************************************}
function TfrmChartTimeSeries.AppendCSVHeader(var MeterBuffer : string) : boolean;
begin
  result := FChartTimeSeries.AppendCSVHeader(MeterBuffer);
end;
{************************************************************************
* NAME:    AppendCSVData
* FUNCTIONAL DESCRIPTION: Append CSV data to string from HMI data stored
*                         in memory resident TimeSeries structure
* INPUT PARAMETERS:       none
* OUTPUT PARAMETERS:      MeterBuffer : string  file content
* RETURN:                 result : boolean   true if file exported, otherwise false
************************************************************************}
function TfrmChartTimeSeries.AppendCSVData(var MeterBuffer : string) : boolean;
begin
  result := FChartTimeSeries.AppendCSVData(MeterBuffer);
end;
{************************************************************************
* NAME:    AppendReport
* FUNCTIONAL DESCRIPTION: Append report summary to string
* INPUT PARAMETERS:    none
* OUTPUT PARAMETERS:   MeterBuffer : string  file content
* RETURN:              result : boolean   true if report appended, otherwise false
************************************************************************}
function TfrmChartTimeSeries.AppendReport(var MeterBuffer : string) : boolean;
begin
  result := FChartTimeSeries.AppendReport(MeterBuffer);
end;
{************************************************************************
* NAME:   FillEventIndexTable
* FUNCTIONAL DESCRIPTION:  Scan raw data, identify events, build index of events
* INPUT PARAMETERS:
*                    filter: boolean  use filter controls true/false
* OUTPUT PARAMETERS: none
* RETURN:  result : boolean  true if events found, otherwise false
************************************************************************}
function TfrmChartTimeSeries.FillEventIndexTable(filter : boolean) : boolean;
begin
  result := FChartTimeSeries.FillEventIndexTable(filter);
end;
{************************************************************************
* NAME:       ReportType
* FUNCTIONAL DESCRIPTION:   return type of report file
* INPUT PARAMETERS: fileData : TStringList   contains content of file
* OUTPUT PARAMETERS:
* RETURN: result : result : TTimeSeriesChartType  (ctSSI, ctLDP, ctUnknown)
************************************************************************}
function TfrmChartTimeSeries.ReportType(filePath : string) : TTimeSeriesChartType;
begin
  result := FChartTimeSeries.EventData.ReportType(filePath);
end;
{************************************************************************
* NAME:    SetChartDateRangeFilters
* FUNCTIONAL DESCRIPTION:
* Set the date range filters based on date/time range in event data
 INPUT PARAMETERS:    none
* OUTPUT PARAMETERS:   none
* RETURN:              none
************************************************************************}
procedure TfrmChartTimeSeries.SetChartDateRangeFilters();
var
  myDateTime1, myDateTime2 : TDateTime;
  dataRecord : TTimeSeriesRecord;
  myFormat : string;
begin
  myFormat := GetDateControlFormat(FDateFormat);
  DateTimePickerEventBeginDate.Format := myFormat;
  DateTimePickerEventEndDate.Format := myFormat;

  if(FChartTimeSeries.EventData.RawDataTable.Count > 0) then
  begin
    dataRecord := (ChartTimeSeries.EventData.RawDataTable[0] as TTimeSeriesRecord);   // first record
    myDateTime1 := dataRecord.DateTime;
    dataRecord := (ChartTimeSeries.EventData.RawDataTable[ChartTimeSeries.EventData.RawDataTable.Count-1] as TTimeSeriesRecord);   // last record
    myDateTime2 := dataRecord.DateTime;

    DateTimePickerEventBeginDate.DateTime := myDateTime1;
    DateTimePickerEventBeginTime.DateTime := myDateTime1;
    DateTimePickerEventEndDate.DateTime := myDateTime2;
    DateTimePickerEventEndTime.DateTime := myDateTime2;
  end else
  begin
    DateTimePickerEventBeginDate.DateTime := Now;
    DateTimePickerEventBeginTime.DateTime := Now;
    DateTimePickerEventEndDate.DateTime := Now;
    DateTimePickerEventEndTime.DateTime := Now;
  end;

end;
{************************************************************************
* NAME:   ChartSeriesActiveState
* FUNCTIONAL DESCRIPTION: Return "active" state of chart series
* INPUT PARAMETERS:  Num : integer  chart series index
* OUTPUT PARAMETERS: none
* RETURN:            result : boolean   active state
************************************************************************}
function TfrmChartTimeSeries.ChartSeriesActiveState(Num : Integer) : boolean;
begin
  result := ChartHMIDisplay.Series[Num].Active
end;
{************************************************************************
* NAME:   GetChartFilename
* FUNCTIONAL DESCRIPTION: Get filename where chart data is saved by default
* INPUT PARAMETERS:  aChartType : THMIChartType (type of chart)
*                    aFileType : TTimeSeriesFileType  (type of file)
*                    ExportPath : default path to file
*                    aID        : ID of meter
* OUTPUT PARAMETERS: none
* RETURN:            filePath : string  path to file that was saved
************************************************************************}
function TfrmChartTimeSeries.GetChartFilename(aChartType : TTimeSeriesChartType; aFileType : TTimeSeriesFileType; const ExportPath, aID : string) : string;
begin
  SetChartType(aChartType);
  result := FChartTimeSeries.GetChartFilename(aFileType, ExportPath, aID);
end;
{************************************************************************
* NAME:     SetChartType
* FUNCTIONAL DESCRIPTION:   Set the chart type
* INPUT PARAMETERS:  aChartType : TTimeSeriesChartType  new chart type
* OUTPUT PARAMETERS: none
* RETURN:            none
************************************************************************}
procedure TfrmChartTimeSeries.SetChartType(aChartType : TTimeSeriesChartType);
begin
  {This function is used to change the chart type based on the incoming
  argument.  The argument indicates the chart type desired.  It is compared
  to the current chart type, and if different, the input settings for the
  current chart object will be saved and the charting data will be freed.
  A new chart object of the requested type will be created and the input
  settings corresponding to the requested chart type will be retrieved and
  made current.
  }
  if(FChartType <> aChartType) then
  begin
    case aChartType of
      ctUndefined,ctLDP:    // need something for Unknown
        begin
          // If input values not saved previously, create a default set for LDP
          if(FChartInputSettingsLDP = nil) then
          begin
            FChartInputSettingsLDP := TTimeSeriesInputSettingsLDP.Create;  // preserved input settings
            FChartInputSettingsLDP.GetChartInputControlSettings;   // Get UI control settings
            FChartInputSettingsLDP.ChartAxisDateTime := true;
          end;
          // InputSettings is a container for the input of the active chart
          if(FChartTimeSeries.EventData.InputSettings = nil) then
          begin
            FChartTimeSeries.EventData.InputSettings := FChartInputSettingsLDP;
          end;

          FChartTimeSeries.SaveChartActiveChannels;  // save settings in previous chart
          FChartTimeSeries.SaveChartActiveEvents;    // save SSI Channels before switching

          FreeAndNil(FChartTimeSeries);
          FChartTimeSeries := TTimeSeriesChartLDP.Create(nil);
          FChartTimeSeries.ChartDisplay := ChartHMIDisplay;
          FChartTimeSeries.CursorTool := ChartTool1;
          FChartTimeSeries.EventData.DateFormat := DateFormat;
          // Make "InputSettings" a reference to the LDP input settings
          FChartTimeSeries.EventData.InputSettings := FChartInputSettingsLDP;
        end;

      ctSSI:
        begin
          // If input values not saved previously, create a default set for SSI
          if(FChartInputSettingsSSI = nil) then
          begin
            FChartInputSettingsSSI := TTimeSeriesInputSettingsSSI.Create;  // preserved input settings
            FChartInputSettingsSSI.GetChartInputControlSettings;  // Get UI control settings
          end;
          // InputSettings is a container for the input of the active chart
          if(FChartTimeSeries.EventData.InputSettings = nil) then
          begin
            FChartTimeSeries.EventData.InputSettings := FChartInputSettingsSSI;
          end;

          FChartTimeSeries.SaveChartActiveChannels;   // save settings in previous chart

          FreeAndNil(FChartTimeSeries);
          FChartTimeSeries := TTimeSeriesChartSSI.Create(nil);
          FChartTimeSeries.ChartDisplay := ChartHMIDisplay;
          FChartTimeSeries.CursorTool := ChartTool1;
          FChartTimeSeries.EventData.DateFormat := DateFormat;
          // Make "InputSettings" a reference to the SSI input settings
          FChartTimeSeries.EventData.InputSettings := FChartInputSettingsSSI;
        end;
      ctInvalid:
      begin
        // do nothing
      end
      else
      begin
        raise Exception.Create(INVALID_CHART_MSG);
      end;
    end;
    FChartType := aChartType;   // now update the chart type
  end;

end;
{************************************************************************
* NAME:       ResetInput
* FUNCTIONAL DESCRIPTION:  Reset chart to match present input settings
* INPUT PARAMETERS: none
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************}
procedure TfrmChartTimeSeries.ResetInput();
begin
  Assert(FChartTimeSeries.EventData <> nil,'EventData is nil');

  with FChartTimeSeries do
  begin
    ChartDisplay.AllowPanning := pmBoth;
    ChartDisplay.AllowZoom := true;
    ChartDisplay.View3D := EventData.InputSettings.Chart3D;
    CursorTool.Active := EventData.InputSettings.ChartQuery;
    CursorTool.Snap := EventData.InputSettings.ChartQuery;
    ChartDisplay.Aspect.Zoom := EventData.InputSettings.AspectZoom;

    with ChartDisplay.View3DOptions do
    begin
      Elevation := EventData.InputSettings.Elevation;
      Rotation := EventData.InputSettings.Rotation;
      Zoom := EventData.InputSettings.Zoom;
      HorizOffset := EventData.InputSettings.HorizOffset;
      VertOffset := EventData.InputSettings.VertOffset;
    end;
    // Restore window position and dimensions
    Width := EventData.InputSettings.WindowWidth;  // these get reset by event handler
    Left := EventData.InputSettings.WindowX;
    Top := EventData.InputSettings.WindowY;
    Height := EventData.InputSettings.WindowHeight;
  end;

end;
{************************************************************************
* NAME:     ResetChartWindow
* FUNCTIONAL DESCRIPTION:   Reset panels of chart window based on chart type
* INPUT PARAMETERS:  none
* OUTPUT PARAMETERS: none
* RETURN:            none
************************************************************************}
procedure TfrmChartTimeSeries.ResetChartWindow;
Const
  TABLIST : array [0..2] of string = ('General','Channels','Events');
  MARGIN_SIZE = 3;
var
  oldTab : integer;
  scaleVisible : boolean;
begin
  GroupBoxEventDateTime.Visible := (ChartType = ctSSI);

  scaleVisible := (FChartType = ctLDP) and
                  (FChartTimeSeries.EventData.LDPChannelTable.Count > 0);

  LabelScale.Visible := scaleVisible;
  ComboBoxScale.Visible := scaleVisible;
  oldTab := FDeviceInfoDialog.TabControlDeviceInfo.TabIndex;  // original tab selection

  GroupBoxEventDateTime.Caption := 'Filter Events';
  if(FChartType = ctSSI) then   // SSI always has chart and event summary
  begin
    ListViewEventList.Width := ScrollBoxEventSelection.Left-MARGIN_SIZE;

    // If panel was previously hidden, alignment must be reset
    PanelEventList.Visible := true;
    PanelEventList.Align := alTop;
    PanelEventList.Height := 207;
    SplitterEventForm.Align := alTop;
    PanelChart.Align := alClient;
    FDeviceInfoDialog.TabControlDeviceInfo.Tabs.Clear;
    FDeviceInfoDialog.TabControlDeviceInfo.Tabs.Add(TABLIST[0]);
    FDeviceInfoDialog.TabControlDeviceInfo.TabIndex := 0;
    TabSetChart.Show;
  end else
  if(ChartType = ctLDP) then   // LDP always has chart, sometimes event summary
  begin
    TabSetChart.Hide;     // hide the SSI tab selector
    if(FChartTimeSeries.EventData.IndexTable.Count > 0)  then  // there are SER events
    begin
      ListViewEventList.Width := PanelEventList.Width - ListViewEventList.Left*2;

      // If panel was previously hidden, alignment must be reset
      PanelEventList.Visible := true;
      PanelEventList.Align := alTop;
      PanelEventList.Height := 175;
      SplitterEventForm.Align := alTop;
      PanelChart.Align := alClient;
    end else
    begin
      PanelEventList.Visible := false;    // no events
    end;

    // display tabs in device info dialog
    
    FDeviceInfoDialog.TabControlDeviceInfo.Tabs.Clear;   // remove all tabs

    // device info
    FDeviceInfoDialog.TabControlDeviceInfo.Tabs.Add(TABLIST[0]);
    if(oldTab = 0) then
    begin
      FDeviceInfoDialog.TabControlDeviceInfo.TabIndex := oldTab;
    end;

    // channel data
    FDeviceInfoDialog.TabControlDeviceInfo.Tabs.Add(TABLIST[1]);
    if(oldTab = 1) then
    begin
      FDeviceInfoDialog.TabControlDeviceInfo.TabIndex := oldTab;
    end;

    // event data
    if(FChartTimeSeries.EventData.LDPEventTable.Count > 0) then  
    begin
      FDeviceInfoDialog.TabControlDeviceInfo.Tabs.Add(TABLIST[2]);
      if(oldTab = 2) then
      begin
        FDeviceInfoDialog.TabControlDeviceInfo.TabIndex := oldTab;
      end;
    end;

  end else  // unknown chart type: display only chart in entire window
  begin
    PanelEventList.Visible := false;
    FDeviceInfoDialog.TabControlDeviceInfo.Tabs.Clear;
    CBEMAChartVisible(false);
  end;
end;
{************************************************************************
* NAME:  TestChartData
* FUNCTIONAL DESCRIPTION: Test contents of chart file for validity
* INPUT PARAMETERS:
*                    aChartType : THMIChartType  type of chart to test data for
*                    filePath : string   path to chart data file
* OUTPUT PARAMETERS:
*                    HasData : boolean  true if header records found, otherwise false
* RETURN:
*                   result : boolean  true if data file was valid, otherwise false
************************************************************************}
function TfrmChartTimeSeries.TestChartData(aChartType : TTimeSeriesChartType; filePath : string; var hasData : boolean) : boolean;
var
  oldChartType : TTimeSeriesChartType;
begin
  oldChartType := FChartType;     // remember original chart type
  result := LoadData(filePath, hasData);  // LoadData might change FChartType
  if(result = false) then    // data was missing or invalid
  begin
    FChartType := oldChartType;
  end;
end;
{************************************************************************
* NAME:   ReadChartData
* FUNCTIONAL DESCRIPTION:  Read HMI data into chart object for display or export
* INPUT PARAMETERS:  aChartType : THMIChartType (type of chart)
*                    fileReset : boolean   True if sending new data,
*                                          False if just showing chart with old data
*                    filePath  : string    path to chart data file
* OUTPUT PARAMETERS:
*                   ChartReset : boolean   true/false initialize chart
*                   ChartValid : boolean   true/false if file has valid data
* RETURN:
*                   result:  true if chartable data found in file, otherwise false
************************************************************************}
function TfrmChartTimeSeries.ReadChartData(aChartType : TTimeSeriesChartType;
              fileReset: boolean; filePath : string;
              var chartReset, chartValid : boolean ) : boolean;
var
  found,hasData : boolean;
  oldChartType,newChartType : TTimeSeriesChartType;
begin
  found := false;  // assume no data is found
  chartValid := true;   // assume if chart data is found it is valid
  chartReset := (fileReset or (ChartType <> ChartTimeSeries.EventData.ChartType));

  oldChartType := FChartType;  // original type

  Assert(FChartTimeSeries.EventData <> nil,'EventData is nil');

  if(chartReset) then
  begin
    newChartType := FChartTimeSeries.EventData.ReadChartType(filePath);   // test file for chart type

    if(newChartType <> ctUndefined) then
    begin
      aChartType := newChartType;   // override incoming chart type
    end;

    case aChartType of
      ctSSI:
      begin
        SetChartType(ctSSI);
        found := LoadData(filePath, hasData);
        if(not found) then
        begin
          if(hasData) then    // there was a valid header but some invalid data
          begin
            chartValid := false;
          end;
        end else
        begin
          ResetInput;
          if(oldChartType <> aChartType) then
          begin
            FChartTimeSeries.EraseChartDisplay;   // clear content of chart
            if(fileReset) then
            begin
              ResetChartDisplay();     // force chart back to 2D un-rotated view
            end;
          end;
          if(fileReset) then
          begin
            SetChartDateRangeFilters;  // reset filters to new range
          end;
          SetChartLabels();
          FChartTimeSeries.FillEventIndexTable(true);  // find events with active filters
          FillEventListViewSSI(false);
          FChartTimeSeries.InitChartVariableCount();
          ChartHMIDisplay.FreeAllSeries();   // throw away previous chart data
          ChartHMICBEMA.FreeAllSeries();   // throw away previous chart data
          if(fileReset) then                   // allocate and initialize channel state
          begin
            FChartTimeSeries.InitChartActiveChannelsDefault(true);
            FChartTimeSeries.InitChartActiveEventsDefault;
          end else
          begin
            FChartTimeSeries.InitChartActiveEvents;
          end;
        end;
      end;

      ctLDP:
      begin
        SetChartType(ctLDP);
        found := LoadData(filePath, hasData);
        if(not found) then
        begin
          if(hasData) then    // there was a valid header but some invalid data
          begin
            chartValid := false;
          end;
        end else
        begin
          ResetInput;
          if(oldChartType <> aChartType) then
          begin
            FChartTimeSeries.EraseChartDisplay;   // clear content of chart
            if(fileReset) then
            begin
              ResetChartDisplay();     // force chart back to 2D un-rotated view
            end;
          end;
          if(fileReset) then
          begin
            SetChartDateRangeFilters;  // reset filters to new range
          end;
          FChartTimeSeries.FillEventIndexTable(true);  // find events with active filters
          FillEventListViewLDP(false);
          FChartTimeSeries.InitChartVariableCount();
          ChartHMIDisplay.FreeAllSeries();   // throw away previous chart data
          ChartHMICBEMA.FreeAllSeries();   // throw away previous chart data
          if(fileReset) then               // allocate and initialize channel state
          begin
            FChartTimeSeries.InitChartActiveChannelsDefault(true);
          end;
        end;
      end;
      else
        begin
          chartValid := false;
          found := false;  // invalid file
        end;
    end;

    // There must be at least 2 records to chart something
    if(FChartTimeSeries.EventData.RawDataTable.Count < 2) then
    begin
      found := false;  // nothing to chart
    end;
  end;

  result := found;
end;
{!******************************************************************************
* MODULE NAME ...  LoadData
* SYSTEM NAME ...  AcSELerator
* AUTHOR ........  Kent Myers
*
* FUNCTIONAL DESCRIPTION:  Read report data from file and insert into data object
*
* INPUT PARAMETERS:
*                    filePath : string  name of file containing ASCII report data
*
* OUTPUT PARAMETERS:
*                    HasData : boolean   true if file has valid header data, otherwise false
* RETURN VALUE:
*                    found : boolean  true if valid data found, otherwise false
*
* ERRORS:
*
* APPLICATION NOTES & EXAMPLE:
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2008.09.24  Kent Myers        Original implementation (obtained from internet)
* 2008.11.17  Kent Myers        Add HasData argument
* 2008.12.04  Kent Myers        Automatically detect file type
*******************************************************************************}
function TfrmChartTimeSeries.LoadData(filePath : string; var hasData : boolean) : boolean;
const
  SKIP_COLUMNS = 3;  // rec, date, time
  STATUS_COLUMN = 4;
var
  fileData : TStringList;
  fileHeader: TStringList;
  fileType : TTimeSeriesFileType;
  found, hasStatus : boolean;
  myHeader, myData : array of integer;
  i, count: Integer;
begin
  fileData := TStringList.Create;
  fileHeader := TStringList.Create;
  try
    filePath := AnsiUpperCase(filePath);

    fileType := ftCSV;   // default value
    FChartTimeSeries.EventData.Clear();    // now it is safe to clear old data

    // read file and get data and labels (header) in string lists

    FChartTimeSeries.EventData.ChartType := FChartType;  // default chart type
    found := ChartTimeSeries.EventData.ReadFileData(filePath, fileData,
                      fileHeader, fileType, myHeader, myData);
    FChartType := ChartTimeSeries.EventData.ChartType; // may have been reset from CSV header data

    hasData := found;  // true if there was valid header data, otherwise false

    // extract raw data from fileData string list and load into time series object
    if(found) then
    begin
      case FChartType of
        ctSSI:     // columns must be previously initialized by calling method
        begin
        end;

        ctLDP:
        begin     // automatically create list of active columns from string list of labels
          if(fileType = ftReport) then  // must initialize data columns
          begin
            // look ahead at column names
            hasStatus := ChartTimeSeries.EventData.ReportFileHasStatusColumn(fileData,count);
            if(hasStatus) then
            begin
              FChartTimeSeries.EventData.HeaderColumns[STATUS_COLUMN] := fileHeader.Count-1;  // status column
              FChartTimeSeries.EventData.SetLengthDataColumns(fileHeader.Count-STATUS_COLUMN);
              // skip first 3 columns (record, date, time)... and last (status)
              for i := SKIP_COLUMNS to fileHeader.Count - 2 do
              begin
                FChartTimeSeries.EventData.DataColumns[i-SKIP_COLUMNS] := i;
              end;
            end else       // generic LDP report
            begin
              FChartTimeSeries.EventData.HeaderColumns[STATUS_COLUMN] := -1;  // no status column
              FChartTimeSeries.EventData.SetLengthDataColumns(fileHeader.Count-SKIP_COLUMNS);
              // skip first 3 columns (record, date, time)
              for i := SKIP_COLUMNS to fileHeader.Count - 1 do
              begin
                FChartTimeSeries.EventData.DataColumns[i-SKIP_COLUMNS] := i;
              end;
            end;
          end;
        end;

        else
          begin
            raise Exception.Create(INVALID_CHART_MSG);
          end;
      end;
      FChartTimeSeries.EventData.LoadData(fileData,
                        FChartTimeSeries.EventData.HeaderColumnArray,
                        FChartTimeSeries.EventData.DataColumnArray);
      FDeviceType := FChartTimeSeries.EventData.DeviceType;
    end;
  finally
    FreeAndNil(fileData);
    FreeAndNil(fileHeader);
  end;

  result := found;
end;
{!******************************************************************************
* MODULE NAME ...  ResetChartControls
* SYSTEM NAME ...  AcSELerator
* AUTHOR ........  Kent Myers
*
* FUNCTIONAL DESCRIPTION:  Reset chart controls based on input settings for current HMI context
*
* INPUT PARAMETERS:  None
*
* OUTPUT PARAMETERS:
*                    None:
* RETURN VALUE:
*                    None:
*
* ERRORS:
*
* APPLICATION NOTES & EXAMPLE:
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2008.09.24  Kent Myers        Original implementation
*******************************************************************************}
procedure TfrmChartTimeSeries.ResetChartControls();
begin
  ComboBoxChartType.ItemIndex := integer(FChartTimeSeries.EventData.InputSettings.ChartStyle);
  SetDateTimeLabels;
end;
{************************************************************************
* NAME:  UpdateActiveEvents
* FUNCTIONAL DESCRIPTION: Update list view with active events
* INPUT PARAMETERS:    none
* OUTPUT PARAMETERS:  none
* RETURN:             none
************************************************************************}
procedure TfrmChartTimeSeries.UpdateActiveEvents;
var
  i : integer;
begin
  for i := 0 to Length(ChartTimeSeries.EventData.InputSettings.ActiveEventArray) -1 do
  begin
    ListViewEventList.Items[i].Checked := ChartTimeSeries.EventData.InputSettings.ActiveEvents[i];
  end;
end;
{************************************************************************
* NAME:  UpdateChart
* FUNCTIONAL DESCRIPTION:  Update chart display after updating chart data
* INPUT PARAMETERS: none
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************}
procedure TfrmChartTimeSeries.UpdateChart;
var
  doCBEMA : boolean;
begin
  doCBEMA := false;
  SetChartDateRangeFilters;  // reset filters to new range
  FChartTimeSeries.FillEventIndexTable(true);  // find events with active filters
  FChartTimeSeries.InitChartVariableCount;   // reset variable count if appropriate
  ChartHMIDisplay.FreeAllSeries();   // throw away previous chart data
  ChartHMICBEMA.FreeAllSeries();   // throw away previous chart data
  FChartTimeSeries.InitChartActiveChannelsDefault(true);   // re-allocate if necessary

  case chartType of
    ctSSI:
    begin
      ChartTimeSeries.EventData.ChartName := 'VSSI';
      ChartTimeSeries.InitChartActiveEventsDefault;
      FillEventListViewSSI(true);
      SetCBEMA(false);
      doCBEMA := true;
    end;
    ctLDP:
    begin
      ChartTimeSeries.EventData.ChartName := 'LDP';
      FillEventListViewLDP(true);
      doCBEMA := false;
    end;
  end;

  DisplayChart(true,true,true,true,doCBEMA); // display the chart
  ButtonImport.Visible := false;

end;
{!******************************************************************************
* MODULE NAME ...  UpdateChart
* SYSTEM NAME ...  AcSELerator
* AUTHOR ........  Kent Myers
*
* FUNCTIONAL DESCRIPTION:  Identify which check box was selected in list view
*
* INPUT PARAMETERS:
*                    reset : boolean  true if regenerating chart, false if updating chart
*
* OUTPUT PARAMETERS:
*                    None:
* RETURN VALUE:
*                    None:
*
* ERRORS:
*
* APPLICATION NOTES & EXAMPLE:
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2008.09.24  Kent Myers        Original implementation
*******************************************************************************}
procedure TfrmChartTimeSeries.UpdateChart(reset : boolean);
var
  doCBEMA : boolean;
begin

  ResetChartWindow;

  doCBEMA := GetCBEMA;

  FChartTimeSeries.SaveChartActiveChannels;   // remember which channels are active
  FChartTimeSeries.InitChartVariableCount;   // reset variable count if appropriate
  FChartTimeSeries.InitChartActiveChannelsDefault(false);   // re-allocate if necessary
  ChartHMIDisplay.FreeAllSeries();   // throw away previous chart data
  ChartHMICBEMA.FreeAllSeries();   // throw away previous chart data
  if(doCBEMA) then
  begin
    DisplayCBEMA(reset);  // regenerate display based on currently checked events
    ChartHMICBEMA.Update();  // refresh display
  end else
  begin
    Display(reset);  // regenerate display based on currently checked events
    ChartHMIDisplay.Update();  // refresh display
  end;

  ResetDeviceInformation;     // update information dialog

end;
{!******************************************************************************
* MODULE NAME ...  ResetChartDisplay
* SYSTEM NAME ...  AcSELerator
* AUTHOR ........  Kent Myers
*
* FUNCTIONAL DESCRIPTION:  Reset chart to the default 2D view
*
* INPUT PARAMETERS:
*                    None:
*
* OUTPUT PARAMETERS:
*                    None:
*
* RETURN VALUE:
*                    None:
*
* ERRORS:
*
* APPLICATION NOTES & EXAMPLE:
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2008.11.20  Kent Myers        Original implementation
*******************************************************************************}
procedure TfrmChartTimeSeries.ResetChartDisplay();
begin

  with FChartTimeSeries.ChartDisplay do
  begin
    AllowPanning := pmBoth;
    AllowZoom := true;
    View3D := false;
    Aspect.Zoom := 100;
  end;

  with FChartTimeSeries.ChartDisplay.View3DOptions do
  begin
    Elevation := 0;
    Rotation := 0;
    Zoom := 100;
    HorizOffset := 0;
    VertOffset := 0;
  end;

  LabelChartToolbar.Caption := '   Drag left button (down/right) to Zoom, Right button to Translate';

  with FChartTimeSeries.EventData.InputSettings do
  begin
    Chart3D := false;
    Elevation := 0;
    Rotation := 0;
    Zoom := 100;
    AspectZoom := 100;
    HorizOffset := 0;
    VertOffset := 0;
  end;
  FChartTimeSeries.AutoScaleAxes;
end;
{************************************************************************
* NAME:   DisplayChart
* FUNCTIONAL DESCRIPTION:  Display Chart data found in report string
* INPUT PARAMETERS:   report : string  is ASCII string containing data to be parsed
* OUTPUT PARAMETERS: none
* RETURN:             result : boolean  true if data could be parsed
************************************************************************}
function TfrmChartTimeSeries.DisplayChart(report : string) : boolean;
var
  filePath : string;
begin
  filePath := GetChartFilename(ctLDP,ftReport,GetSpecialFolder(CSIDL_PERSONAL)+'\',' ');
  SaveStringToFile(filePath,report);
  result := DisplayChartFile(filePath);
  DeleteFile(filePath);
end;
{************************************************************************
* NAME:   DisplayChartFile
* FUNCTIONAL DESCRIPTION:  Display Chart data found in file
* INPUT PARAMETERS:   filePath : filename containing data to be parsed
* OUTPUT PARAMETERS: none
* RETURN:             result : boolean  true if data could be parsed
************************************************************************}
function TfrmChartTimeSeries.DisplayChartFile(filePath : string) : boolean;
var
  myChartType : TTimeSeriesChartType;
  myFileType : TTimeSeriesFileType;
begin
  myChartType := ChartTimeSeries.EventData.ReadChartType(filePath,myFileType);
  SetChartType(myChartType);
  InitChart(myFileType);
  result := DisplayChart(myChartType,true,filePath,true);
end;
{************************************************************************
* NAME:   DisplayChart
* FUNCTIONAL DESCRIPTION:  Display Chart data found in file
* INPUT PARAMETERS:   aChartType : THMIChartType  type of chart to display
*                     fileReset : boolean  Initialize display of chart (true/false)
*                     filename : string  is the original filename
*                     showChart : boolean  display the chart (true/false)
* OUTPUT PARAMETERS: none
* RETURN:             result : boolean  true if data could be parsed
************************************************************************}
function TfrmChartTimeSeries.DisplayChart(aChartType : TTimeSeriesChartType; fileReset: boolean; filePath : string; showChart : boolean ) : boolean;
var
  chartReset, chartValid, found, doCBEMA : boolean;
  oldChartType : TTimeSeriesChartType;
  myInputSettings : TTimeSeriesInputSettingsSSI;
begin
  oldChartType := FChartType;  // original type
  chartReset := true;

  if(fileReset) then    // cancel date/time filter controls
  begin
    CheckBoxEventBegin.Checked := false;
    CheckBoxEventEnd.Checked := false;
    if(FChartType = ctSSI) then
    begin
      myInputSettings := (ChartTimeSeries.EventData.InputSettings as TTimeSeriesInputSettingsSSI);
      myInputSettings.BeginDateActive := false;
      myInputSettings.EndDateActive := false;
    end;
  end;

  doCBEMA := GetCBEMA;
  SetCBEMA(false);       // must disable CBEMA so that VSSI chart will be generated

  // Read chart data from file  (chartReset true if fileReset or change chart type)
  found := ReadChartData(aChartType, fileReset, filePath, chartReset, chartValid);
  if(not chartValid) then
  begin
    ShowMessage('Error reading data into chart.');
  end;

  DisplayChart(found,chartReset,showChart,fileReset,doCBEMA);

  if(found or (not chartReset and (FChartTimeSeries.EventData.RawDataTable.Count > 1))) then
  begin
  // successful chart
  end else
  begin      // something went wrong
    FChartType := oldChartType;  // restore original chart type
  end;

  result := found;
end;
{************************************************************************
* NAME:   DisplayChart
* FUNCTIONAL DESCRIPTION:  Display Chart data already loaded into memory
* INPUT PARAMETERS:   found : boolean   data is in memory
*                     chartReset : boolean true fileReset or ChartType changed
*                     fileReset : boolean  true if new file imported
*                     showChart : boolean  display the chart (true/false)
*                     doCBEMA : boolean true if CBEMA is default chart
* OUTPUT PARAMETERS: none
* RETURN:             none
************************************************************************}
procedure TfrmChartTimeSeries.DisplayChart(found,chartReset,showChart,fileReset,doCBEMA : boolean);
begin
  if(found) then      // valid charting data found in file
  begin
    UpdateChart(true);   // update the chart
    UpdateActiveEvents;  // update list view with active events
  end;

  if(found or (not chartReset and (FChartTimeSeries.EventData.RawDataTable.Count > 1))) then
  begin
    if(showChart or Visible) then           // Display the chart window
    begin
      WindowState := wsNormal;   // cancel window maximization
      Show;
      if((FChartType = ctSSI) and (fileReset or doCBEMA)) then
      begin
        if(TabSetChart.TabIndex <> 0) then  // VSSI chart is active
        begin
          TabSetChart.TabIndex := 0;   // force CBEMA chart to be active
        end else
        begin                         // CBEMA tab active, draw CBEMA explicitly
          CBEMAChartVisible(doCBEMA);  // Make CBEMA chart visible
          DisplayCBEMA(true);  // regenerate CBEMA chart
          ChartHMICBEMA.Update();  // refresh display
        end;
      end;
    end else
    begin
      SetCBEMA(doCBEMA);  // Set CBEMA flag in case it was disabled.
    end;
  end;
end;
{!******************************************************************************
* MODULE NAME ...  Display
* SYSTEM NAME ...  AcSELerator
* AUTHOR ........  Kent Myers
*                       SetChartVariableCount
* FUNCTIONAL DESCRIPTION:  Create the chart, assign attributes and insert data for display
*
* INPUT PARAMETERS:
*                    reset: boolean  true if regenerating display (rescale axis limits),
*                                    false if updating display (retain current axis limits)
* OUTPUT PARAMETERS:
*                    None:
* RETURN VALUE:
*                    None:
*
* ERRORS:
*
* APPLICATION NOTES & EXAMPLE:
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2008.09.24  Kent Myers        Original implementation
* 2008.12.24  Kent Myers        Fix event indicators when they overlap
*******************************************************************************}
procedure TfrmChartTimeSeries.Display(reset : boolean);
const
  MARKERS_PER_EVENT = 4;
  SKIP_RECORD = '(5)';
  TEST_RECORD = '(6)';
var
  index,i,j,checkCount,channelCount,rec : integer;
  myRecord,xMin,xMax,yMin,yMax : double;
  dataRecord        : TTimeSeriesRecord;
  channelRecord     : TTimeSeriesLDPChannelRecord;
  dataIndex         : TTimeSeriesIndexRecord;
  start, stop, estart, estop : integer;
  time_format : string;
  channelScale : array of single;
  parentEvent, activeCount,aCount,markCount : integer;
  chartMarker,chartRec,eventRec,eventNumber : array of integer;   // event number of charted events
  chartHeight : array of single;
  eventColor : array of TColor;
begin
  if(ChartTimeSeries.ChartVariableCount = 0) then exit;   // no data

  CBEMAChartVisible(false);

  // Event marker information (possible 4 markers per event)
  SetLength(chartRec,FChartTimeSeries.EventData.IndexTable.Count);  // sequential record number
  SetLength(eventRec,FChartTimeSeries.EventData.IndexTable.Count);  // actual record number
  SetLength(chartMarker,FChartTimeSeries.EventData.IndexTable.Count*MARKERS_PER_EVENT); // seq. record for marker
  SetLength(eventNumber,FChartTimeSeries.EventData.IndexTable.Count*MARKERS_PER_EVENT); // event # of marker
  SetLength(eventColor,FChartTimeSeries.EventData.IndexTable.Count*MARKERS_PER_EVENT); // color of marker
  SetLength(chartHeight,FChartTimeSeries.EventData.IndexTable.Count*MARKERS_PER_EVENT); // height of marker

  checkCount := GetActiveEventCount;

  // Initialize channel scaling
  channelCount := FChartTimeSeries.EventData.RawDataLabels.Count;  // number of channels to display
  SetLength(channelScale,channelCount);    // scale factors for channels
  for j := 0 to channelCount - 1 do      // assign default scaling values
  begin
    channelScale[j] := 1.0;
  end;

  case FChartType of
    ctSSI:
    begin
      time_format := DATETIME_WITH_MS;    // display milliseconds
    end;
    ctLDP:
    begin
      time_format := DATETIME_WITH_SEC;
      if(not ChartTimeSeries.EventData.InputSettings.ChartSecondaryUnits) then
      begin
        for j := 0 to ChartTimeSeries.EventData.LDPChannelTable.Count - 1 do
        begin
          channelRecord := (ChartTimeSeries.EventData.LDPChannelTable[j] as TTimeSeriesLDPChannelRecord);
          channelScale[j] := channelRecord.ChannelScale;
        end;
      end;
    end;
  end;

  SetLength(ChartSeries,FChartTimeSeries.ChartVariableCount);

  // Initialize chart axis scale

  InitChartMinMax(reset,xMin,xMax,yMin,yMax);

  // Initialize chart attributes

  InitChartAttributes(channelCount);

  // Fill data into chart series

  if((FChartType = ctLDP) or (FChartTimeSeries.EventData.IndexTable.Count = 0)) then     // no events, display all data
  begin
    for index := 0 to FChartTimeSeries.EventData.RawDataTable.Count-1 do
    begin
      dataRecord := (FChartTimeSeries.EventData.RawDataTable[index] as TTimeSeriesRecord);
      if((dataRecord.Status = SKIP_RECORD) or (dataRecord.Status = TEST_RECORD)) then
      begin
        continue;  // don't display empty records
      end;
      myRecord := dataRecord.RecordNumber;
      for j := 0 to channelCount-1 do     // add every channel for this record to the chart
      begin
        FChartTimeSeries.ChartDisplay.Series[j].AddXY(myRecord,dataRecord.Value[j]*channelScale[j]);
      end;
    end;

    if(FChartType = ctLDP) then // display SER events
    begin
      DisplaySEREvents(checkCount);
    end;

  end else   // display only events in chart
  begin
    if(FChartTimeSeries.ChartHasEvents) then
    begin
      BarSeries.SeriesColor := INTERVAL_COLOR;
    end;

    rec := 0;
    activeCount := 0;
    markCount := 0;
    for i := 0 to FChartTimeSeries.EventData.IndexTable.Count-1 do   // look at event index records
    begin
      dataIndex := (FChartTimeSeries.EventData.IndexTable[i] as TTimeSeriesIndexRecord); // get the event index record
      if((checkCount > 0) and (dataindex.Checked = false)) then continue;   // event not selected for display

      start := dataIndex.FirstRecord; // get first/last record for complete event
      stop := dataIndex.LastRecord;   //      including pre-/post-disturbance
      estart := dataIndex.StartEvent; // get first/last record for actual event
      estop := dataIndex.StopEvent;   //      excluding pre-/post-disturbance

      parentEvent := -1;  // assume no sub-events
      if(FChartType = ctSSI) then // test for sub-event only if SSI
      begin
        aCount := 0;  // active events so far
        for j := 0 to i-1 do
        begin
          dataIndex := (FChartTimeSeries.EventData.IndexTable[j] as TTimeSeriesIndexRecord);
          if((checkCount > 0) and not dataIndex.Checked) then continue;
          if(start < dataIndex.LastRecord) then   // it is a sub-event of active event
          begin
            parentEvent := aCount;   // index of parent event
            break;
          end;
          Inc(aCount);
        end;
      end;

      chartRec[activeCount] := rec;
      eventRec[activeCount] := start;

      for index := start to stop do
      begin
        if(parentEvent >= 0) then   // this is a sub-event (don't display)
        begin
          myRecord := index - eventRec[parentEvent] + chartRec[parentEvent]+1;
        end else
        begin              // standard event, display data trace
          myRecord := rec+1;
          dataRecord := (ChartTimeSeries.EventData.RawDataTable[index] as TTimeSeriesRecord);
          for j := 0 to channelCount-1 do    // add every channel for this record to the chart
          begin
            FChartTimeSeries.ChartDisplay.Series[j].AddXY(myRecord,dataRecord.Value[j]);
          end;
        end;

        RecordEventMarkers(i,parentEvent,index,start,stop,estart,estop,myRecord,
                           chartMarker,eventNumber,chartHeight,eventColor,markCount);

        if(parentEvent < 0) then  // only increment record if not sub-event
        begin
          inc(rec);
        end;
      end;
      Inc(activeCount);
    end;

    // Display event markers

    DisplayEventMarkers(checkCount,markCount,chartMarker,eventNumber,
                        chartHeight,eventColor);
  end;

  // Update display attributes

  FChartTimeSeries.InitChartTitles;
  FChartTimeSeries.InitChartActiveChannels;
  SetChartLineStyles;
  if(reset) then
  begin
    ResetChartControls;   // update UI controls from present settings
  end;
  UpdateChartMinMax(reset,xMin,xMax,yMin,yMax);

end;
{************************************************************************
* NAME:   GetActiveEventCount
* FUNCTIONAL DESCRIPTION: Get number of active events in the chart
* INPUT PARAMETERS: none
* OUTPUT PARAMETERS: none
* RETURN: result : integer  number of active events
************************************************************************}
function TfrmChartTimeSeries.GetActiveEventCount : integer;
var
  checkCount,j : integer;
  dataIndex         : TTimeSeriesIndexRecord;
begin
  checkCount := 0;    // initialize number events selected for display (SSI only)
  for j := 0 to FChartTimeSeries.EventData.IndexTable.Count-1 do   // count selected events
  begin
    dataIndex := (FChartTimeSeries.EventData.IndexTable[j] as TTimeSeriesIndexRecord); // get the event index record
    if(dataIndex.Checked) then
    begin
      inc(checkCount);
    end;
  end;
  result := checkCount;
end;
{************************************************************************
* NAME:   GetCBEMA
* FUNCTIONAL DESCRIPTION: Get the CBEMA option from the input settings object
* INPUT PARAMETERS:
* OUTPUT PARAMETERS:
* RETURN: result : boolean  true if CBEMA active otherwise false
************************************************************************}
function TfrmChartTimeSeries.GetCBEMA : boolean;
begin
  if(FChartType = ctSSI) then
  begin
    result := (ChartTimeSeries.EventData.InputSettings as TTimeSeriesInputSettingsSSI).ChartCBEMA;
  end else
  begin
    result := false;
  end;
end;
{************************************************************************
* NAME:   SetCBEMA
* FUNCTIONAL DESCRIPTION: Set the CBEMA option in the input settings object
* INPUT PARAMETERS: doCBEMA : boolean  true if CBEMA chart display is desired
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************}
procedure TfrmChartTimeSeries.SetCBEMA(doCBEMA : boolean);
begin
  if(FChartType = ctSSI) then
  begin
    (FChartTimeSeries.EventData.InputSettings as TTimeSeriesInputSettingsSSI).ChartCBEMA := doCBEMA;
  end;
end;
{************************************************************************
* NAME:  InitChartMinMax
* FUNCTIONAL DESCRIPTION:  initialize chart axis scaling
* INPUT PARAMETERS: reset : boolean  true if discarding current scale information
* OUTPUT PARAMETERS:
*                   xMin : single  minimum value of x-axis scale
*                   xMax : single  maximum value of x-axis scale
*                   yMin : single  minimum value of y-axis scale
*                   yMax : single  maximum value of y-axis scale
* RETURN: none
************************************************************************}
procedure TfrmChartTimeSeries.InitChartMinMax(reset : boolean; var xMin,xMax,yMin,yMax : double);
begin
  if(reset) then   // throw away current scale information
  begin
    xMin := 0.0;
    xMax := 1.0;
    yMin := 0.0;
    yMax := 1.0;
  end else
  begin           // remember current axis limits
    xMin := FChartTimeSeries.ChartDisplay.BottomAxis.Minimum;
    xMax := FChartTimeSeries.ChartDisplay.BottomAxis.Maximum;
    yMin := FChartTimeSeries.ChartDisplay.LeftAxis.Minimum;
    yMax := FChartTimeSeries.ChartDisplay.LeftAxis.Maximum;
  end;
  FChartTimeSeries.ChartDisplay.BottomAxis.Automatic := false;
  FChartTimeSeries.ChartDisplay.LeftAxis.Automatic := false;

end;
{************************************************************************
* NAME:    UpdateChartMinMax
* FUNCTIONAL DESCRIPTION:   Update chart axis scaling after chart is finished
* INPUT PARAMETERS: reset : boolean  true if discarding current scale information
*                   xMin : single  minimum value of x-axis scale
*                   xMax : single  maximum value of x-axis scale
*                   yMin : single  minimum value of y-axis scale
*                   yMax : single  maximum value of y-axis scale
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************}
procedure TfrmChartTimeSeries.UpdateChartMinMax(reset : boolean; xMin,xMax,yMin,yMax : double);
begin
  if(reset) then
  begin
    FChartTimeSeries.AutoScaleAxes;   // use customized axis scaling
  end else
  begin
    FChartTimeSeries.ChartDisplay.BottomAxis.Automatic := false;
    FChartTimeSeries.ChartDisplay.LeftAxis.Automatic := false;
    FChartTimeSeries.ChartDisplay.BottomAxis.Minimum := xMin;
    FChartTimeSeries.ChartDisplay.BottomAxis.Maximum := xMax;
    FChartTimeSeries.ChartDisplay.LeftAxis.Minimum := yMin;
    FChartTimeSeries.ChartDisplay.LeftAxis.Maximum := yMax;
  end;
end;
{************************************************************************
* NAME:   InitChartAttributes
* FUNCTIONAL DESCRIPTION:   Initialize attributes of the chart
* INPUT PARAMETERS: channelCount : integer   number of indep. variables
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************}
procedure TfrmChartTimeSeries.InitChartAttributes(channelCount : integer);
const
  CBEMA_TITLES : array [0..1] of string = ('ITI Sag','ITI Swell');
  EVENT_TITLE = 'Event';
var
  j : integer;
begin
  FChartTimeSeries.ChartDisplay.Title.Font.Color := clBlack;
  FChartTimeSeries.ChartDisplay.Title.Font.Name := 'Tahoma';
  FChartTimeSeries.ChartDisplay.Title.Font.Size := 12;
  FChartTimeSeries.ChartDisplay.BottomAxis.RoundFirstLabel := True;
  FChartTimeSeries.ChartDisplay.Legend.CheckBoxes := true;
  FChartTimeSeries.ChartDisplay.Legend.Title.Visible := true;
  FChartTimeSeries.ChartDisplay.Legend.Title.Text.Clear;
  FChartTimeSeries.ChartDisplay.Legend.Title.Text.Add('Channel');

  FChartTimeSeries.ChartDisplay.FreeAllSeries();   // throw away previous chart data

  if(FChartTimeSeries.EventData.InputSettings.ChartAxisDateTime) then    // display date/time X axis labels
  begin
    With FChartTimeSeries.ChartDisplay.BottomAxis do
    begin
      LabelsAngle := 90;
    end;
  end else                               // display record number X axis labels
  begin
    With ChartTimeSeries.ChartDisplay.BottomAxis do
    begin
      LabelsAngle := 0;
      Automatic := true;
    end;
  end;

  // Allocate and initialize chart series

  for j := 0 to FChartTimeSeries.ChartVariableCount-1 do        // for each channel
  begin
    case FChartTimeSeries.EventData.InputSettings.ChartStyle of
      sctLine:
        begin
          ChartSeries[j] := TLineSeries.Create(self);    // create line chart
        end;
      sctArea:
        begin
          ChartSeries[j] := TAreaSeries.Create(self);   // create area chart
          TAreaSeries(ChartSeries[j]).Transparency := FChartTimeSeries.EventData.InputSettings.ChartTransparency;  // percent transparency
        end;
      else
      begin
        Assert(false,'Invalid chart style.');
      end;
    end;

    if(FChartTimeSeries.EventData.RawDataTable.Count > 50) then
    begin
      ChartSeries[j].AreaLinesPen.Visible := false;   // suppress grid lines
    end else
      ChartSeries[j].AreaLinesPen.Visible := true;


    if(j >= channelCount) then  // it is a CBEMA curve
    begin
      Assert(((j-channelCount) >= 0) and
              ((j-channelCount) < Length(CBEMA_TITLES)-1),
              'Array length mismatch');
      ChartSeries[j].Stairs := true;  // use step function style for CBEMA
      ChartSeries[j].Title := CBEMA_TITLES[j-channelCount];  // assign channel name to legend
    end else
    begin
      ChartSeries[j].Title := ChartTimeSeries.EventData.RawDataLabels[j];  // assign channel name to legend
    end;

    FChartTimeSeries.ChartDisplay.AddSeries(ChartSeries[j]);           // add the new channel series to the chart
  end;

  if((FChartType = ctSSI) and FChartTimeSeries.ChartHasEvents) then
  begin
    BarSeries := TBarSeries.Create(self);
    BarSeries.Title := EVENT_TITLE;  // assign channel name to legend
    Barseries.Marks.Visible := ChartTimeSeries.EventData.InputSettings.ChartLabels;
    Barseries.Marks.Clip := true;
    Barseries.Marks.Shadow.Clip := true;
    BarSeries.Marks.Shadow.Visible := false;   // hide shadow (doesn't clip)
    FChartTimeSeries.ChartDisplay.AddSeries(BarSeries);  // add the new series to the chart
  end;

  if((ChartType = ctLDP) and ChartTimeSeries.EventsInInterval) then
  begin
    PointSeries := TPointSeries.Create(self);
    PointSeries.Title := EVENT_TITLE;  // assign channel name to legend
    PointSeries.Marks.Visible := ChartTimeSeries.EventData.InputSettings.ChartLabels;
    PointSeries.Marks.Clip := true;
    PointSeries.Marks.Shadow.Clip := true;
    PointSeries.Marks.Arrowlength := 20;
    PointSeries.Marks.Shadow.Visible := false;   // hide shadow (doesn't clip)
    FChartTimeSeries.ChartDisplay.AddSeries(PointSeries);  // add the new series to the chart
  end;
end;
{************************************************************************
* NAME:    DisplaySEREvents
* FUNCTIONAL DESCRIPTION:   Display LDP events in chart
* INPUT PARAMETERS: checkCount : number of active channels
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************}
procedure TfrmChartTimeSeries.DisplaySEREvents(checkCount : integer);
var
  value,yvalue : double;
  rec,i : integer;
  myLabel : string;
  dataIndex         : TTimeSeriesIndexRecord;
begin
  // display all event symbols along a line where LDP chart data is minimum
  yvalue := FChartTimeSeries.ChartDisplay.MinYValue(FChartTimeSeries.ChartDisplay.LeftAxis);
  rec := 0;
  for i := 0 to FChartTimeSeries.EventData.IndexTable.Count - 1 do
  begin
    dataIndex := (FChartTimeSeries.EventData.IndexTable[i] as TTimeSeriesIndexRecord);
    // See if event occurred within range of LDP date/time limits
    value := FChartTimeSeries.GetRecordFromEventTime(dataIndex.DateTime);
    if(value < 0.0) then  // not in range of LDP date/time limits
    begin
      continue;
    end;
    if((checkCount > 0) and not dataIndex.Checked) then  // not selected by user
    begin
      continue;
    end;
    // trim off the leading # sign
    myLabel := AnsiRightStr(dataIndex.EventName,Length(dataIndex.EventName)-1);
    PointSeries.AddXY(value,yvalue); // value is interpolated record number
    PointSeries.ValueColor[rec] := EVENTMARKER_COLOR;
    PointSeries.Marks[rec].Text.Add(myLabel);  // marker is event number
    Inc(rec);
  end;
end;
{************************************************************************
* NAME:    RecordEventMarkers
* FUNCTIONAL DESCRIPTION:   Record information about SSI event markers 
* INPUT PARAMETERS: iEvent : integer event number
*                   parentEvent : integer  event number of parent event
*                   index : integer  record number
*                   start : integer  beginning of pre-disturbance interval
*                   stop : integer   end of post-disturbance interval
*                   estart : integer beginning of actual event
*                   estop  : integer end of actual event
* OUTPUT PARAMETERS:
*                   chartMarker : array of integer  rec where marker appears
*                   eventNumber : array of integer  event number label of marker
*                   chartHeight : array of single   height of marker
*                   eventColor : array of TColor   color of marker
*                   markCount : integer number of markers
* RETURN: none
************************************************************************}
procedure TfrmChartTimeSeries.RecordEventMarkers(iEvent, parentEvent: integer;
                             index,start,stop,estart,estop : integer;
                             myRecord : double;
                             var chartMarker,eventNumber : array of integer;
                             var chartHeight : array of single;
                             var eventColor : array of TColor;
                             var markCount : integer);
const
  INTERVAL_HEIGHT = 100.0;
  EVENT_HEIGHT = 50.0;
begin
  if(FChartTimeSeries.chartHasEvents) then
  begin
    if((index = start) or (index = stop)) then     // pre/post event interval
    begin
      chartHeight[markCount] := INTERVAL_HEIGHT;
      if((index = estart) or (index = estop)) then  // also limit of actual event
      begin
        eventColor[markCount] := PARENTEVENT_COLOR;   // default color
        if(parentEvent >= 0) then       // sub-event
        begin
          eventColor[markCount] := SUBEVENT_COLOR;
          chartHeight[markCount] := EVENT_HEIGHT;
        end;
        if(index = estart) then
        begin
          eventNumber[markCount] := -(iEvent+1);   // neg. indicates beginning
        end else
        begin
          eventNumber[markCount] := iEvent+1;      // pos. indicates end
        end;
        chartMarker[markCount] := Trunc(myRecord);
        Inc(markCount);
      end else
      begin
        eventNumber[markCount] := 0; // don't display label
        chartMarker[markCount] := Trunc(myRecord);
        eventColor[markCount] := INTERVAL_COLOR;
        Inc(markCount);
      end;
    end else
    if((index = estart) or (index = estop)) then    // Must end with interrupt sub-event
    begin
      chartHeight[markCount] := EVENT_HEIGHT;
      if(index = estart) then
      begin
        eventNumber[markCount] := -(iEvent+1);   // neg. indicates beginning
      end else
      begin
        eventNumber[markCount] := iEvent+1;     // pos. indicates end
      end;
      chartMarker[markCount] := Trunc(myRecord);
      eventColor[markCount] := PARENTEVENT_COLOR;
      Inc(markCount);
    end;
  end;
end;
{************************************************************************
* NAME:    DisplayEventMarkers
* FUNCTIONAL DESCRIPTION:  Display SSI event markers in chart
* INPUT PARAMETERS: checkCount : integer number of active channels in legend
*                   markCount : integer number of markers
*                   chartMarker : array of integer  rec where marker appears
*                   eventNumber : array of integer  event number label of marker
*                   chartHeight : array of single   height of marker
*                   eventColor : array of TColor   color of marker
* OUTPUT PARAMETERS: (if used)
* RETURN: (if not a procedure)
************************************************************************}
procedure TfrmChartTimeSeries.DisplayEventMarkers(checkCount,markCount : integer;
                             chartMarker,eventNumber : array of integer;
                             chartHeight : array of single;
                             eventColor : array of TColor);
const
  EN_MAX = 5;   // maximum number of coincident event markers
var
  rec, start, stop : integer;
  i, j, index,lastMarker : integer;
  dataIndex         : TTimeSeriesIndexRecord;
  ivalue : integer;
  icolor : TColor;
  rvalue : single;
  enList : array [0..5] of integer;   // list of coincident markers
  enCount,enFound : integer;
  myLabel : string;
begin
{Event markers identify the beginning and end of each event.  There are two
markers to delimit the entire event including pre-disturbance and post-disturbance.
There are two markers to delimit the actual event excluding the pre-/post-disturbance
intervals.  Sometimes there is no pre- or post-disturbance interval.

Sorting is required in case an event interval contains one or more sub-events.
If this is the case, the sub-events have been merged into the parent event
and the terminating event has been removed.   The terminating marker from
the parent event must follow the markers for the sub-events (interrupts).
}
  // sort by event markers by record number where event ends

  for i := 0 to markCount - 2 do
  begin
    for j := i+1 to markCount - 1 do
    begin
      if(chartMarker[i] > chartMarker[j]) then
      begin
        ivalue := chartMarker[j];
        chartMarker[j] := chartMarker[i];
        chartMarker[i] := ivalue;

        ivalue := eventNumber[j];
        eventNumber[j] := eventNumber[i];
        eventNumber[i] := ivalue;

        icolor := eventColor[j];
        eventColor[j] := eventColor[i];
        eventColor[i] := icolor;

        rvalue := chartHeight[j];
        chartHeight[j] := chartHeight[i];
        chartHeight[i] := rvalue;
      end;
    end;
  end;

  rec := 0;
  stop := 0;  // use to test for sub-events
  lastMarker := 0;  // last marker drawn, don't repeat search of previously drawn markers
  for i := 0 to FChartTimeSeries.EventData.IndexTable.Count-1 do   // look at event index records
  begin
    dataIndex := (FChartTimeSeries.EventData.IndexTable[i] as TTimeSeriesIndexRecord); // get the event index record
    if((checkCount > 0) and (dataIndex.Checked = false)) then continue;   // event not selected for display
    if(dataIndex.LastRecord <= stop) then continue;  // sub-event, skip it
    start := dataIndex.FirstRecord; // get first/last record for this event
    stop := dataIndex.LastRecord;
    enFound := 0;    // position in marker list where marker matches record
    // NOTE: index is just a counter, it is not used inside this for loop
    for index := start to stop+1 do   // range of records for this event
    begin
      enCount := 0;                   // count number of coincident markers
      for j := lastMarker to markCount - 1 do   // See if this record should be marked
      begin
        if(rec < chartMarker[j]) then break;   // no need to keep looking at markers
        if(rec = chartMarker[j]) then  // display marker for this record
        begin
          if(enCount < EN_MAX) then  // make list of coincident markers
          begin
            enList[enCount] := eventNumber[j];
            enFound := j;       // use this position to get height, color
            lastMarker := j+1;  // next marker to start search
            Inc(enCount);       // count number of coincident markers
          end;
        end else
      end;

      if(enCount > 0) then         // display event markers
      begin
        BarSeries.AddXY(rec,chartHeight[enFound]);
        BarSeries.ValueColor[rec] := eventColor[enFound];

        // NOTE: pos/neg event number may be used to distinguish start/end of event

        myLabel := '';
        for j := 0 to enCount - 1 do // create a label for up to 5 concident markers
        begin
          if(enList[j] < 0) then   // beginning of event interval
          begin
            myLabel := myLabel + Format('%d,',[-enList[j]]);
          end else                 // end of event interval
          if(enList[j] > 0) then
          begin
            myLabel := myLabel + Format('%d,',[enList[j]]);
          end;
        end;

        if(Length(myLabel) > 0) then   // add a label to the marker
        begin
          myLabel := AnsiLeftStr(myLabel,Length(myLabel)-1);  // trim trailing comma
          BarSeries.Marks[rec].Text.Add(myLabel);
        end else
        begin
          BarSeries.Marks[rec].Text.Add('');  // no label
        end;
      end else
      begin                 // add placeholder markers
        BarSeries.AddXY(rec,0.0);
        BarSeries.Marks[rec].Text.Add('');
        BarSeries.ValueColor[rec] := INTERVAL_COLOR;
      end;
      Inc(rec);
    end;
  end;
end;
{!******************************************************************************
* MODULE NAME ...  DisplayCBEMA
* SYSTEM NAME ...  AcSELerator
* AUTHOR ........  Kent Myers
* FUNCTIONAL DESCRIPTION:  Create the chart, assign attributes and insert data
*                          display of CBEMA curves
*
* INPUT PARAMETERS:
*                    reset: boolean  true if regenerating display (rescale axis limits),
*                                    false if updating display (retain current axis limits)
* OUTPUT PARAMETERS:
*                    None:
* RETURN VALUE:
*                    None:
*
* ERRORS:
*
* APPLICATION NOTES & EXAMPLE:
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2008.09.24  Kent Myers        Original implementation
*******************************************************************************}
procedure TfrmChartTimeSeries.DisplayCBEMA(reset : boolean);
const
  CBEMA_COLOR : array[0..7] of TColor =
  (
    clGreen, clOlive, clWebGreenYellow, clYellow, clWebBrown, clWebDarkOrange, clWebOrangeRed, clRed
  );
var
  i,checkCount,rec,start,stop : integer;
  xMin,xMax,yMin,yMax : double;
  duration, amplitude : double;
  dataIndex         : TTimeSeriesIndexRecord;
  myTimeStamp : TTimeStamp;
  sDuration, sAmplitude : array of double;
  sEvent : array of integer;
  sEventType : array of TTimeSeriesCBEMAViolationType;
  sagValues, swellValues : array of double;
  maxDuration : double;
begin
  if(FChartTimeSeries.ChartVariableCount = 0) then exit;   // no data

  maxDuration := 30.0;   // initialize to minimum reasonable value

  checkCount := GetActiveEventCount;

  InitChartMinMax(reset,xMin,xMax,yMin,yMax);

  InitChartAttributesCBEMA;

  SetLength(sDuration,FChartTimeSeries.EventData.IndexTable.Count);
  SetLength(sEvent,FChartTimeSeries.EventData.IndexTable.Count);
  SetLength(sEventType,FChartTimeSeries.EventData.IndexTable.Count);
  SetLength(sAmplitude,FChartTimeSeries.EventData.IndexTable.Count);

  rec := 0;
  for i := 0 to FChartTimeSeries.EventData.IndexTable.Count-1 do   // look at event index records
  begin
    dataIndex := (FChartTimeSeries.EventData.IndexTable[i] as TTimeSeriesIndexRecord); // get the event index record
    if((checkCount > 0) and (dataindex.Checked = false)) then continue;   // event not selected for display

    start := dataIndex.FirstRecord; // get first/last record for this event
    stop := dataIndex.LastRecord;

    SetLength(sagValues,stop-start+1);
    SetLength(swellValues,stop-start+1);
    sEventType[rec] := dataIndex.Violation;

    // Convert a TDateTime duration to milliseconds

    myTimeStamp := DateTimeToTimeStamp(dataIndex.Duration);
    // get number of days since reference date
    Dec(myTimeStamp.Date, TTimeStamp(DateTimeToTimeStamp(0)).Date);
    // number of milliseconds
    duration := (myTimeStamp.Date * 24 * 60 * 60) + (myTimeStamp.Time / 1000.0);
    amplitude := dataIndex.Magnitude;
    sDuration[rec] := duration;
    sAmplitude[rec] := amplitude;
    if(maxDuration < duration) then
    begin
      maxDuration := duration;
    end;
    sEvent[rec] := i+1;
    inc(rec);
  end;

  SortEventsByDuration(rec,sDuration,sAmplitude,sEvent,sEventType);

  for i := 0 to rec-1 do   // Update event marker color/number
  begin
    FChartTimeSeries.ChartDisplay.Series[0].AddXY(sDuration[i],sAmplitude[i]);
    FChartTimeSeries.ChartDisplay.Series[0].Marks[i].Text.Add(Format('%d',[sEvent[i]]));
    FChartTimeSeries.ChartDisplay.Series[0].ValueColor[i] := CBEMA_COLOR[integer(SEventType[i])];
  end;

  // Display Sag/Swell limits

  DisplayCBEMALimits(maxDuration);

  // Update display attributes  (titles and scaling)

  FChartTimeSeries.InitChartTitles;
  With FChartTimeSeries.ChartDisplay.BottomAxis do
  begin
    Increment := 0.001;
    Minimum := 0.0;
    AutomaticMinimum := false;
    LabelsAngle := 0;
    Automatic := true;
  end;

  UpdateChartMinMax(reset,xMin,xMax,yMin,yMax);
end;
{************************************************************************
* NAME:       InitChartAttributesCBEMA
* FUNCTIONAL DESCRIPTION:   Initialize chart attributes for CBEMA chart
* INPUT PARAMETERS: none
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************}
procedure TfrmChartTimeSeries.InitChartAttributesCBEMA;
const
  CBEMA_LABELS : array [0..2] of string = ('Sag Limit','Swell Limit','Event');
begin
  FChartTimeSeries.ChartDisplay.Title.Font.Color := clBlack;
  FChartTimeSeries.ChartDisplay.Title.Font.Name := 'Tahoma';
  FChartTimeSeries.ChartDisplay.Title.Font.Size := 12;
  FChartTimeSeries.ChartDisplay.BottomAxis.RoundFirstLabel := True;
  FChartTimeSeries.ChartDisplay.Legend.CheckBoxes := true;
  FChartTimeSeries.ChartDisplay.Legend.Title.Visible := true;
  FChartTimeSeries.ChartDisplay.Legend.Title.Text.Clear;
  FChartTimeSeries.ChartDisplay.Legend.Title.Text.Add('Legend');

  FChartTimeSeries.ChartDisplay.FreeAllSeries();   // throw away previous chart data

  // Allocate and initialize event series

  CBEMADataSeries := TPointSeries.Create(self);    // create line chart
  CBEMADataSeries.Title := CBEMA_LABELS[2];  // assign channel name to legend
  CBEMADataSeries.Pointer.Style := psCircle;
  CBEMADataSeries.Pointer.HorizSize := 5;
  CBEMADataSeries.Pointer.VertSize := 5;
  CBEMADataSeries.Pointer.Visible := true;
  CBEMADataSeries.Marks.Visible := FChartTimeSeries.EventData.InputSettings.ChartLabels;
  CBEMADataSeries.Marks.Clip := true;
  CBEMADataSeries.Marks.Shadow.Clip := true;
  CBEMADataSeries.Marks.Shadow.Visible := false;  // hide shadow (doesn't clip)
  CBEMADataSeries.Marks.ArrowLength := 20;
  FChartTimeSeries.ChartDisplay.AddSeries(CBEMADataSeries);  // add the new channel series to the chart
end;
{************************************************************************
* NAME:          SortEventsByDuration
* FUNCTIONAL DESCRIPTION:  Sort events by duration
* INPUT PARAMETERS: eventCount : integer number of events
* OUTPUT PARAMETERS:
*                   sDuration : array of double   event duration
*                   sAmplitude : array of double   event amplitude
*                   sEvent  : array of integer    event number
*                   sEventType : TTimeSeriesCBEMAViolation  violation type
* RETURN: none
************************************************************************}
procedure TfrmChartTimeSeries.SortEventsByDuration(eventCount : integer;
                               var sDuration, sAmplitude : array of double;
                               var sEvent : array of integer;
                               var sEventType : array of TTimeSeriesCBEMAViolationType);
var
  i,j,event : integer;
  myEventType : TTimeSeriesCBEMAViolationType;
  duration, amplitude : double;
begin
  for i := 0 to eventCount - 2 do   // sort events by duration
  begin
    for j := (i+1) to eventCount - 1 do
    begin
       if(sDuration[i] > sDuration[j]) then
       begin
         duration := sDuration[i];
         sDuration[i] := sDuration[j];
         sDuration[j] := duration;
         amplitude := sAmplitude[i];
         sAmplitude[i] := sAmplitude[j];
         sAmplitude[j] := amplitude;
         event := sEvent[i];
         sEvent[i] := sEvent[j];
         sEvent[j] := event;
         myEventType := sEventType[i];
         SEventType[i] := sEventType[j];
         SEventType[j] := myEventType;
       end;
    end;
  end;
end;
{************************************************************************
* NAME:  DisplayCBEMALimits
* FUNCTIONAL DESCRIPTION:  Display the sag and swell limit curves
* INPUT PARAMETERS: (if used)
* OUTPUT PARAMETERS: (if used)
* RETURN: (if not a procedure)
************************************************************************}
procedure TfrmChartTimeSeries.DisplayCBEMALimits(maxDuration : double);
const
  CBEMA_PT_LABELS : array [0..2] of string = ('Swell Limit','Sag Limit','Event');
  CBEMA_SAG_DURATION : array [0..7] of double = (0.0199, 0.02,0.5,10.0,15.0,20.0,25.0,10000.0);
  CBEMA_SAG_AMPLITUDE : array [0..7] of double = (0.0,70.0,80.0,90.0,90.0,90.0,90.0,90.0);
  CBEMA_SWELL_DURATION : array [0..8] of double = (0.0001,0.001,0.003,0.0031,0.5,1.0,10.0,20.0,10000.0);
  CBEMA_SWELL_AMPLITUDE : array [0..8] of double = (500.0,200.0,140.0,120.0, 110.0, 110.0, 110.0, 110.0, 110.0);
var
  i : integer;
  myCBEMA_SAG_DURATION : array [0..7] of double;
  myCBEMA_SWELL_DURATION : array [0..8] of double;
begin
  for i := 0 to 1 do  // create each CBEMA sag/swell limit curve
  begin
    CBEMALimitSeries[i] := TLineSeries.Create(self);
    CBEMALimitSeries[i].Title := CBEMA_PT_LABELS[i];  // assign channel name to legend
    CBEMALimitSeries[i].Stairs := true;  // use step function style for CBEMA
    CBEMALimitSeries[i].Visible := true;  // display step curves
    CBEMALimitSeries[i].LinePen.Width := 2;
    FChartTimeSeries.ChartDisplay.AddSeries(CBEMALimitSeries[i]); // add the new channel series to the chart
  end;

  for i := Low(CBEMA_SAG_DURATION) to High(CBEMA_SAG_DURATION)-1 do
  begin
    myCBEMA_SAG_DURATION[i] := CBEMA_SAG_DURATION[i];
  end;
  myCBEMA_SAG_DURATION[High(CBEMA_SAG_DURATION)] := maxDuration;

  for i := Low(CBEMA_SWELL_DURATION) to High(CBEMA_SWELL_DURATION)-1 do
  begin
    myCBEMA_SWELL_DURATION[i] := CBEMA_SWELL_DURATION[i];
  end;
  myCBEMA_SWELL_DURATION[High(CBEMA_SWELL_DURATION)] := maxDuration;

  for i := Low(CBEMA_SWELL_DURATION) to High(CBEMA_SWELL_DURATION) do
  begin
    FChartTimeSeries.ChartDisplay.Series[1].AddXY(myCBEMA_SWELL_DURATION[i],CBEMA_SWELL_AMPLITUDE[i]);
  end;

  for i := Low(CBEMA_SAG_DURATION) to High(CBEMA_SAG_DURATION) do
  begin
    FChartTimeSeries.ChartDisplay.Series[2].AddXY(myCBEMA_SAG_DURATION[i],CBEMA_SAG_AMPLITUDE[i]);
  end;
end;
{!******************************************************************************
* MODULE NAME ...  SetDateTimeLabels
* SYSTEM NAME ...  AcSELerator
* AUTHOR ........  Kent Myers
*
* FUNCTIONAL DESCRIPTION:  set chart date/time labels
*
* INPUT PARAMETERS:
*                    Sender : TObject
*
* OUTPUT PARAMETERS:
*                    None:
* RETURN VALUE:
*                    None:
*
* ERRORS:
*
* APPLICATION NOTES & EXAMPLE:
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2008.11.13  Kent Myers        Original implementation
*******************************************************************************}
procedure TfrmChartTimeSeries.SetDateTimeLabels();
const
  DATE_PT_LENGTH_LDP = 110;
  DATE_PT_LENGTH_SSI = 125;
  RECORD_PT_LENGTH = 15;
begin
  with FChartTimeSeries do
  begin

    if((EventData.ChartType = ctSSI) and     // Time/date not used for CBEMA
       TTimeSeriesInputSettingsSSI(EventData.InputSettings).ChartCBEMA) then
    begin
      exit;
    end;

    if(EventData.InputSettings.ChartAxisDateTime) then
    begin
      ChartDisplay.BottomAxis.LabelsAngle := 90;
      if(EventData.ChartType = ctSSI) then
      begin
        ChartDisplay.BottomAxis.LabelsSize := DATE_PT_LENGTH_SSI;
      end else
      begin
        ChartDisplay.BottomAxis.LabelsSize := DATE_PT_LENGTH_LDP;
      end;
    end else
    begin
      ChartDisplay.BottomAxis.LabelsAngle := 0;
      ChartDisplay.BottomAxis.LabelsSize := RECORD_PT_LENGTH;
    end;
    InitChartTitles();
    ChartDisplay.Update();
  end;
end;
{!******************************************************************************
* MODULE NAME ...  SetChartLineStyles
* SYSTEM NAME ...  AcSELerator
* AUTHOR ........  Kent Myers
*
* FUNCTIONAL DESCRIPTION:  Set chart line styles (color, type)
*
* INPUT PARAMETERS:
*                    None:
*
* OUTPUT PARAMETERS:
*                    None:
*
* RETURN VALUE:
*                    None:
*
* ERRORS:
*
* APPLICATION NOTES & EXAMPLE:
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2008.10.1  Kent Myers        Original implementation
*******************************************************************************}
procedure TfrmChartTimeSeries.SetChartLineStyles();
var
  i , pen_width: integer;
  cPhase : THMIChannelPhase;
  cType : THMIChannelType;
begin

  if(FChartTimeSeries.ChartVariableCount = 0) then exit;   // no data
  if(FChartTimeSeries.EventData.ChartType = ctSSI) then
  begin
    if(TTimeSeriesInputSettingsSSI(FChartTimeSeries.EventData.InputSettings).ChartCBEMA) then
    begin
      exit;
    end;
  end;

  if(Length(ChartSeries) < 1) then   // no data loaded into chart
  begin
    exit;
  end;

  // choose default pen widths.  Some channels will use different pen widths.
  if(FChartTimeSeries.ChartDisplay.View3D) then
  begin
    pen_width := 1;    // want thin lines for 3D
  end else
  begin
    pen_width := 3;    // want thick lines for 2D  (voltages and other channels)
  end;

  // update pen width for each channel
  for i := 0 to FChartTimeSeries.EventData.RawDataLabels.Count-1 do
  begin
    cPhase := FChartTimeSeries.GetChannelPhase(ChartTimeSeries.EventData.RawDataLabels[i]);
    cType := FChartTimeSeries.GetChannelType(ChartTimeSeries.EventData.RawDataLabels[i]);
    ChartSeries[i].LinePen.Width := pen_width;
    ChartSeries[i].SeriesColor := FChartTimeSeries.ChartLineColor(cPhase,i); // assign color to channel

    case FChartTimeSeries.EventData.InputSettings.ChartStyle of
      sctLine:
        begin
          if(cType = ctCurrent) then    // channel is current so use different pen width
          begin
            if(FChartTimeSeries.ChartDisplay.View3D) then
            begin
              ChartSeries[i].LinePen.Width := 1;    // want thin lines for 3D
            end else
            begin
              ChartSeries[i].LinePen.Width := 2;
            end;
          end;
          ChartSeries[i].AreaLinesPen.Style := FChartTimeSeries.ChartLineStyle(cType);
          ChartSeries[i].LinePen.Style := FChartTimeSeries.ChartLineStyle(cType);
        end;
      sctArea:
        begin
        end;
      else
        begin
          Assert(false,'Invalid chart style.');
        end;
    end;
  end;
end;
{!******************************************************************************
* MODULE NAME ...  SetChartLabels
* SYSTEM NAME ...  AcSELerator
* AUTHOR ........  Kent Myers
*
* FUNCTIONAL DESCRIPTION:
* Set the chart variable names that are Form 5/9 specific
*
* INPUT PARAMETERS:
*                    None:
* OUTPUT PARAMETERS:
*                    None:
*
* RETURN VALUE:
*                   None:
*
* ERRORS:
*
* APPLICATION NOTES & EXAMPLE:
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2008.11.06  Kent Myers        Original code
*******************************************************************************}
procedure TfrmChartTimeSeries.SetChartLabels();
const
  FIRST_VOLTAGE_4XX = 0;
  FIRST_VOLTAGE = 5;
var
  voltageColumn : integer;    // skip currents
begin
  Assert(FChartTimeSeries.EventData <> nil,'EventData is nil');

  if(FChartTimeSeries.EventData.RawDataLabels.Count < 7) then   // SEL4XX
  begin
    voltageColumn := FIRST_VOLTAGE_4XX;
  end else
  begin
    voltageColumn := FIRST_VOLTAGE;
  end;

  case FChartType of
    ctSSI:              // update SSI chart variable names
      begin
        if(AnsiCompareText(FChartTimeSeries.EventData.RawDataLabels[voltageColumn],'VAB') = 0) then
        begin
          ComboBoxEventPhase.Items[0] := 'AB';
          ComboBoxEventPhase.Items[1] := 'BC';
          ComboBoxEventPhase.Items[2] := 'CA';
        end else
        begin
          ComboBoxEventPhase.Items[0] := 'A';
          ComboBoxEventPhase.Items[1] := 'B';
          ComboBoxEventPhase.Items[2] := 'C';
        end;
      end;
    ctLDP:
      begin
      end;
    else
      begin
        Assert(False, 'unknown chart type');
      end;
  end;
end;
{************************************************************************
* NAME:    CBEMAChartVisible
* FUNCTIONAL DESCRIPTION: select type of chart to display for SSI
* INPUT PARAMETERS: doCBEMA : boolean  true if CBEMA chart is visible
* OUTPUT PARAMETERS: none
* RETURN: none     result : boolean   true if display should be refreshed
************************************************************************}
function TfrmChartTimeSeries.CBEMAChartVisible(doCBEMA : boolean) : boolean;
const
  SSI_TAB = 1;
var
  reset : boolean;
begin
  Assert(FChartTimeSeries.EventData <> nil,'EventData is nil');

  FChartTimeSeries.EventData.InputSettings.ChartQuery := false;    // cancel cursor query
  labelChartType.enabled := not doCBEMA;
  comboboxChartType.enabled := not doCBEMA;
  speedButtonDateTime.visible := not doCBEMA;
  toolbar1.Realign;

  if(doCBEMA) then    // CBEMA
  begin
    TTimeSeriesInputSettingsSSI(FChartTimeSeries.EventData.InputSettings).ChartCBEMA := true;
    ChartHMIDisplay.SendToBack;
    reset := (ChartHMICBEMA.SeriesCount = 0);
    FChartTimeSeries.ChartDisplay := ChartHMICBEMA;   // set active chart pointer
    FChartTimeSeries.CursorTool := ChartTool2;
    labelChartTransparency.Enabled := false;
    scrollbarChartTransparency.Enabled := false;
  end else              // SSI/LDP
  begin
    TTimeSeriesInputSettingsSSI(FChartTimeSeries.EventData.InputSettings).ChartCBEMA := false;
    ChartHMICBEMA.SendToBack;
    reset := false;  // no need to reset
    FChartTimeSeries.ChartDisplay := ChartHMIDisplay;   // set active chart pointer
    FChartTimeSeries.CursorTool := ChartTool1;
    labelChartTransparency.Enabled := (comboboxChartType.ItemIndex = SSI_TAB);
    scrollbarChartTransparency.Enabled := (comboboxChartType.ItemIndex = SSI_TAB);
  end;
  FChartTimeSeries.CursorTool.Active := false;
  FChartTimeSeries.CursorTool.Snap := true;
  result := reset;
end;
{************************************************************************
* NAME:   FillEventListViewLDP
* FUNCTIONAL DESCRIPTION:  Fill list with LDP event information
* INPUT PARAMETERS: aReset : boolean  true will initialize checked entries
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************}

procedure TfrmChartTimeSeries.FillEventListViewLDP(aReset : boolean);
const
  NO_EVENT_MSG = 'Status: No events were found within the LDP data acquisition interval.';
  EVENT_MSG = 'Status: %d events found.';
  COLUMN_COUNT = 4;
  COLUMNSIZE : array [0..COLUMN_COUNT-1] of integer =
     (60,  80, 80, 350);
  COLUMN_NAMES : array [0..COLUMN_COUNT-1] of string =
     ('Event','Date','Time','Description');
  // Event,Date,Start,Description
var
  i : integer;
  indexRecord : TTimeSeriesIndexRecord;
  listItem : TListItem;
  myDateTime : string;
  doRestore : boolean;
  checkList,selectList : array of boolean;
  myColumns : array [0..COLUMN_COUNT-1] of integer;
  myFormat : TFormatSettings;
begin
  FChartTimeSeries.EventData.InitDateTimeFormat(ftInternational,myFormat);

  Assert(FChartTimeSeries.EventData <> nil,'EventData is nil');

  if(FChartTimeSeries.EventData.IndexTable.Count = 0) then
  begin
    with ListViewEventList do
    begin
      Clear();
      Columns.Clear();    // Erase existing columns
    end;
    LabelStatus.Caption := NO_EVENT_MSG;
    exit;  // no data
  end;

  // preserve state if count is the same
  doRestore := (FChartTimeSeries.EventData.IndexTable.Count = ListViewEventList.Items.Count);
  if(doRestore) then
  begin
    SetLength(checkList,ListViewEventList.Items.Count);
    SetLength(selectList,ListViewEventList.Items.Count);
    for i := 0 to ListViewEventList.Items.Count - 1 do
    begin
      selectList[i] := ListViewEventList.Items[i].Selected;
      checkList[i] := ListViewEventList.Items[i].Checked;
    end;
  end;

  // Restore previous column width if rows present
  if((ListViewEventList.Columns.Count = COLUMN_COUNT) and
     (ListViewEventList.Items.Count > 0)) then
  begin
    for i := 0 to COLUMN_COUNT-1 do   // restore previous columns widths
    begin
      myColumns[i] :=   ListViewEventList.Columns[i].Width;
    end;
  end else
    for i := 0 to COLUMN_COUNT-1 do   // set default column width
    begin
      myColumns[i] :=   COLUMNSIZE[i];
    end;
  begin

  end;

  with ListViewEventList do
  begin
    Clear();
    Columns.Clear();    // Erase existing columns
  end;

  with ListViewEventList do
  begin
    for i := 0 to COLUMN_COUNT-1 do
    begin
      Columns.Add;
      Columns[i].Width := myColumns[i];
      Columns[i].AutoSize := false;
      Columns[i].Caption := COLUMN_NAMES[i];
    end;

    // fill list view with contents of event index

    for i := 0 to FChartTimeSeries.EventData.IndexTable.Count-1 do
    begin
      indexRecord := (FChartTimeSeries.EventData.IndexTable[i] as TTimeSeriesIndexRecord); // get the event index record
      listItem := Items.Add;
      listItem.Caption := indexRecord.EventName;
      listItem.Checked := false;
      myDateTime := DateToStr(IndexRecord.DateTime,myFormat);  // setting date format
      listItem.SubItems.Add(myDateTime);
      DateTimeToString(myDateTime,TIME_WITH_MS,indexRecord.DateTime);
      listItem.SubItems.Add(myDateTime);
      listItem.SubItems.Add(indexRecord.Description);
    end;

    // Update status label based on number of events found in index

    LabelStatus.Caption := Format(EVENT_MSG,[ChartTimeSeries.EventData.IndexTable.Count]);
  end;

  // Restore selected item(s)
  if(doRestore and not aReset) then
  begin
    for i := 0 to ListViewEventList.Items.Count - 1 do
    begin
      ListViewEventList.Items[i].Checked := checkList[i];
      ListViewEventList.Items[i].Selected := selectList[i];
      ListViewEventList.Items[i].Focused := selectList[i];
    end;
  end;

end;
{************************************************************************
* NAME:     FillEventListViewSSI
* FUNCTIONAL DESCRIPTION:  Create list of events in UI List View control
* INPUT PARAMETERS: aReset : boolean  true will initialize checked entries
* OUTPUT PARAMETERS: none
* RETURN:           none
************************************************************************}
procedure TfrmChartTimeSeries.FillEventListViewSSI(aReset : boolean);
const
  NO_EVENTS_MSG = 'Status: No events were found matching the present event selection criteria.';
  EVENTS_FOUND_FMT = 'Status: %d events found.';
  EVENTS_FOUND_PHASE_FMT = 'Status: %d events found in which %s on Phase %s occurred.';
  EVENTS_FOUND_TYPE_FMT = 'Status: %d events found in which %s occurred.';
  EVENTS_FOUND_MIXED_FMT = 'Status: %d events found in which a Sag, Swell, or Interrupt on Phase %s occurred.';
  COLUMN_COUNT = 13;
  DATAFORMAT = '%5.1f';
  MYPHASE : array [0..3] of string = ('A','B','C','Any');
  MYTYPE : array [-1..3] of string = ('None','a Sag','a Swell','an Interrupt','Any');
  COLUMNSIZE : array [0..COLUMN_COUNT-1] of integer =
     (60,   45,   70,   80,    85,       80,        75,     60,   60,   60,   60,   60,   60);
  // Event, Type, Date, Start, Duration, Magnitude, Region, VaMin...
  COLUMN_NAMES : array [0..COLUMN_COUNT-1] of string =
     ('Event','Type','Date','Time','Duration','Event Depth','ITIC Region',
      'Va Min','Va Max','Vb Min','Vb Max','Vc Min','Vc Max');
var
  i : integer;
  indexRecord : TTimeSeriesIndexRecord;
  listItem : TListItem;
  myDateTime,eventMag : string;
  inputSettings : TTimeSeriesInputSettingsSSI;
  myColumns : array [0..COLUMN_COUNT-1] of integer;
  myFormat : TFormatSettings;
  doRestore : boolean;
  checkList,selectList : array of boolean;
begin
  FChartTimeSeries.EventData.InitDateTimeFormat(ftInternational,myFormat);

  if(FChartTimeSeries.EventData.IndexTable.Count = 0) then
  begin
    with ListViewEventList do
    begin
      Clear();
      Columns.Clear();    // Erase existing columns
    end;
    LabelStatus.Caption := NO_EVENTS_MSG;
    exit;  // no data
  end;

  // preserve state if count is the same
  doRestore := (FChartTimeSeries.EventData.IndexTable.Count = ListViewEventList.Items.Count);
  if(doRestore) then
  begin
    SetLength(checkList,ListViewEventList.Items.Count);
    SetLength(selectList,ListViewEventList.Items.Count);
    for i := 0 to ListViewEventList.Items.Count - 1 do
    begin
      selectList[i] := ListViewEventList.Items[i].Selected;
      checkList[i] := ListViewEventList.Items[i].Checked;
    end;
  end;

  // Restore previous column width if rows present
  if((ListViewEventList.Columns.Count = COLUMN_COUNT) and
     (ListViewEventList.Items.Count > 0)) then
  begin
    for i := 0 to COLUMN_COUNT-1 do   // restore previous columns widths
    begin
      myColumns[i] :=   ListViewEventList.Columns[i].Width;
    end;
  end else
  begin
    for i := 0 to COLUMN_COUNT-1 do   // set default column width
    begin
      myColumns[i] :=   COLUMNSIZE[i];
    end;
  end;

  inputSettings := (FChartTimeSeries.EventData.InputSettings as TTimeSeriesInputSettingsSSI);

  with ListViewEventList do
  begin
    Clear();
    Columns.Clear();    // Erase existing columns
  end;

  with ListViewEventList do
  begin
    for i := 0 to COLUMN_COUNT-1 do
    begin
      Columns.Add;
      Columns[i].Width := myColumns[i];
      Columns[i].AutoSize := false;
      Columns[i].Caption := COLUMN_NAMES[i];
    end;

    // fill list view with contents of event index

    for i := 0 to FChartTimeSeries.EventData.IndexTable.Count-1 do
    begin
      indexRecord := (FChartTimeSeries.EventData.IndexTable[i] as TTimeSeriesIndexRecord); // get the event index record
      listItem := Items.Add;
      listItem.Caption := indexRecord.EventName;
      listItem.Checked := false;
      listItem.SubItems.Add(indexRecord.Description);
      myDateTime := DateToStr(indexRecord.DateTime,myFormat);  // setting date format
      listItem.SubItems.Add(myDateTime);
      DateTimeToString(myDateTime,TIME_WITH_MS,indexRecord.DateTime);
      listItem.SubItems.Add(myDateTime);
      DurationToTimeString(myDateTime,indexRecord.Duration);
      listItem.SubItems.Add(myDateTime);
      if(indexRecord.EventType = evNone) then    // Trigger event
      begin
        eventMag := '$$$$$';    // Magnitude not meaningful
      end else
      begin
        eventMag := Format(DATAFORMAT,[indexRecord.Magnitude]);
      end;
      listItem.SubItems.Add(eventMag);
      listItem.SubItems.Add(ChartTimeSeries.CBEMAViolationString(indexRecord.Violation));
      listItem.SubItems.Add(Format(DATAFORMAT,[indexRecord.VaMin]));
      listItem.SubItems.Add(Format(DATAFORMAT,[indexRecord.VaMax]));
      listItem.SubItems.Add(Format(DATAFORMAT,[indexRecord.VbMin]));
      listItem.SubItems.Add(Format(DATAFORMAT,[indexRecord.VbMax]));
      listItem.SubItems.Add(Format(DATAFORMAT,[indexRecord.VcMin]));
      listItem.SubItems.Add(Format(DATAFORMAT,[indexRecord.VcMax]));
    end;

    // Update status label based on number of events found in index

    if((InputSettings.EventPhase = epAll) and (InputSettings.EventType = evAll)) then
    begin
      LabelStatus.Caption := Format(EVENTS_FOUND_FMT,
                [FChartTimeSeries.EventData.IndexTable.Count]);
    end else
    begin
      if((InputSettings.EventPhase <> epAll) and (InputSettings.EventType <> evAll)) then
      begin
        LabelStatus.Caption := Format(EVENTS_FOUND_PHASE_FMT,
                 [FChartTimeSeries.EventData.IndexTable.Count,
                 MYTYPE[integer(inputSettings.EventType)],
                 MYPHASE[integer(inputSettings.EventPhase)]]);
      end else
      if(InputSettings.EventPhase = epAll) then
      begin
        LabelStatus.Caption := Format(EVENTS_FOUND_TYPE_FMT,
                 [FChartTimeSeries.EventData.IndexTable.Count,
                 MYTYPE[integer(inputSettings.EventType)],
                 MYPHASE[integer(inputSettings.EventPhase)]]);
      end else
      begin
        LabelStatus.Caption := Format(EVENTS_FOUND_MIXED_FMT,
                 [FChartTimeSeries.EventData.IndexTable.Count,
                 MYPHASE[integer(inputSettings.EventPhase)]]);
      end;
    end;
  end;

  // Restore selected item(s)
  if(doRestore and not aReset) then
  begin
    for i := 0 to ListViewEventList.Items.Count - 1 do
    begin
      ListViewEventList.Items[i].Checked := checkList[i];
      ListViewEventList.Items[i].Selected := selectList[i];
      ListViewEventList.Items[i].Focused := selectList[i];
    end;
  end;

end;
{************************************************************************
* NAME:    ShowDeviceInformation
* FUNCTIONAL DESCRIPTION: Display information about the device
* INPUT PARAMETERS:       none
* OUTPUT PARAMETERS:      none
* RETURN:                 none
************************************************************************}
procedure TfrmChartTimeSeries.ShowDeviceInformation;
const
  COLUMN_COUNT = 3;
  COLUMNSIZE : array [0..COLUMN_COUNT-1] of integer = (50,120,350);
  COLUMN_NAMES : array [0..COLUMN_COUNT-1] of string =
    ('Item','Property','Value');
var
  i : integer;
  listItem : TListItem;
  myColumns : array [0..COLUMN_COUNT-1] of integer;
  settingRecord : TTimeSeriesSettingRecord;
  myFormat : TFormatSettings;
  myDate, myTime : string;
begin
  ChartTimeSeries.EventData.InitDateTimeFormat(ftInternational,myFormat);

  myDate := DateToStr(ChartTimeSeries.EventData.ChartDateTime,myFormat);  // setting date format
  myTime := FormatDateTime('hh:mm:ss',ChartTimeSeries.EventData.ChartDateTime);
  // Restore previous column width if rows present
  with FDeviceInfoDialog.ListViewDeviceInfo do
  begin
    if(Items.Count > 0) then
    begin
      for i := 0 to COLUMN_COUNT-1 do   // restore previous columns widths
      begin
        myColumns[i] :=   Columns[i].Width;
      end;
    end else
    begin
      for i := 0 to COLUMN_COUNT-1 do   // set default column width
      begin
        myColumns[i] :=   COLUMNSIZE[i];
      end;
    end;

    Clear();
    Columns.Clear();    // Erase existing columns

    for i := 0 to COLUMN_COUNT-1 do
    begin
      Columns.Add;
      Columns[i].Width := myColumns[i];
      Columns[i].AutoSize := false;
      Columns[i].Caption := COLUMN_NAMES[i];
    end;

    // fill list view with contents of device information

    listItem := Items.Add;
    listItem.Caption := ' 1';
    listItem.SubItems.Add('Chart Type');
    listItem.SubItems.Add(ChartTimeSeries.EventData.ChartName);

    listItem := Items.Add;
    listItem.Caption := ' 2';
    listItem.SubItems.Add('Date');
    listItem.SubItems.Add(myDate);

    listItem := Items.Add;
    listItem.Caption := ' 3';
    listItem.SubItems.Add('Time');
    listItem.SubItems.Add(myTime);

    listItem := Items.Add;
    listItem.Caption := ' 4';
    listItem.SubItems.Add('Time Source');
    listItem.SubItems.Add(ChartTimeSeries.EventData.TimeSource);

    listItem := Items.Add;
    listItem.Caption := ' 5';
    listItem.SubItems.Add('Device ID');
    listItem.SubItems.Add(ChartTimeSeries.EventData.MeterID);

    listItem := Items.Add;
    listItem.Caption := ' 6';
    listItem.SubItems.Add('Device FID');
    listItem.SubItems.Add(ChartTimeSeries.EventData.MeterFID);

    listItem := Items.Add;
    listItem.Caption := ' 7';
    listItem.SubItems.Add('Device TID');
    listItem.SubItems.Add(ChartTimeSeries.EventData.MeterTID);

    for i := 0 to FChartTimeSeries.EventData.SettingTable.Count - 1 do
    begin
      listItem := Items.Add;
      listItem.Caption := Format(' %d',[Items.Count]);
      settingRecord := (ChartTimeSeries.EventData.SettingTable[i] as TTimeSeriesSettingRecord);
      listItem.SubItems.Add(settingRecord.SettingName);
      listItem.SubItems.Add(Format('%.2f',[settingRecord.SettingValue]));
    end;
  end;
end;
{************************************************************************
* NAME:    ShowChannelInformation
* FUNCTIONAL DESCRIPTION: Display channel information about the device
* INPUT PARAMETERS:       none
* OUTPUT PARAMETERS:      none
* RETURN:                 none
************************************************************************}
procedure TfrmChartTimeSeries.ShowChannelInformation;
const
  COLUMN_COUNT = 5;
  COLUMNSIZE : array [0..COLUMN_COUNT-1] of integer = (120,100,80,110,130);
  COLUMN_NAMES : array [0..COLUMN_COUNT-1] of string =
      ('Name','Function','Units','Scaling Rule','Scaling Factor');
var
  i : integer;
  listItem : TListItem;
  convertDecimal : boolean;
  myColumns : array [0..COLUMN_COUNT-1] of integer;
  channelRecord : TTimeSeriesLDPChannelRecord;
begin

  if(FChartTimeSeries.EventData.LDPChannelTable.Count = 0) then
  begin
    FDeviceInfoDialog.ListViewChannelInfo.Clear();
    FDeviceInfoDialog.ListViewChannelInfo.Columns.Clear();    // Erase existing columns
    exit;  // no data
  end;

  convertDecimal := AnsiContainsStr(Format('%.1f',[1.0]),',');  // test string for decimal delimiter

  // Restore previous column width if rows present
  with FDeviceInfoDialog.ListViewChannelInfo do
  begin
    if((Items.Count > 0) and 
       (Columns.Count = COLUMN_COUNT)) then
    begin
      for i := 0 to COLUMN_COUNT-1 do   // restore previous columns widths
      begin
        myColumns[i] :=   Columns[i].Width;
      end;
    end else
    begin
      for i := 0 to COLUMN_COUNT-1 do   // set default column width
      begin
        myColumns[i] :=   COLUMNSIZE[i];
      end;
    end;

    Clear();
    Columns.Clear();    // Erase existing columns

    for i := 0 to COLUMN_COUNT-1 do
    begin
      Columns.Add;
      Columns[i].Width := myColumns[i];
      Columns[i].AutoSize := false;
      Columns[i].Caption := COLUMN_NAMES[i];
    end;

    for i := 0 to FChartTimeSeries.EventData.LDPChannelTable.Count - 1 do
    begin
      channelRecord := (FChartTimeSeries.EventData.LDPChannelTable[i] as TTimeSeriesLDPChannelRecord); // get the event index record
      ListItem := Items.Add;
      ListItem.Caption := channelRecord.ChannelName;
      ListItem.SubItems.Add(channelRecord.ChannelFunction);
      ListItem.SubItems.Add(channelRecord.ChannelUnits);
      if(convertDecimal) then
      begin
        ListItem.SubItems.Add(AnsiReplaceStr(channelRecord.ChannelRules,'.',','));
      end else
      begin
        ListItem.SubItems.Add(channelRecord.ChannelRules);
      end;
      ListItem.SubItems.Add(Format('%.2f',[channelRecord.ChannelScale]));
    end;
  end;
end;
{************************************************************************
* NAME:    ShowEventInformation
* FUNCTIONAL DESCRIPTION: Display information about the SER events
* INPUT PARAMETERS:       none
* OUTPUT PARAMETERS:      none
* RETURN:                 none
************************************************************************}
procedure TfrmChartTimeSeries.ShowEventInformation;
const
  COLUMN_COUNT = 5;
  COLUMNSIZE : array [0..COLUMN_COUNT-1] of integer = (60,100,120,350,60);
  COLUMN_NAMES : array [0..COLUMN_COUNT-1] of string =
       ('Event','Date','Time','Description','Code');
var
  i : integer;
  listItem : TListItem;
  myColumns : array [0..COLUMN_COUNT-1] of integer;
  eventRecord : TTimeSeriesLDPEventRecord;
  myDateTime : string;
  myFormat : TFormatSettings;
begin
  FChartTimeSeries.EventData.InitDateTimeFormat(ftInternational,myFormat);

  if(FChartTimeSeries.EventData.LDPEventTable.Count = 0) then
  begin
    FDeviceInfoDialog.ListViewEventInfo.Clear();
    FDeviceInfoDialog.ListViewEventInfo.Columns.Clear();    // Erase existing columns
    exit;  // no data
  end;

  // Restore previous column width if rows present
  with FDeviceInfoDialog.ListViewEventInfo do
  begin
    if((Items.Count > 0) and 
       (Columns.Count = COLUMN_COUNT)) then
    begin
      for i := 0 to COLUMN_COUNT-1 do   // restore previous columns widths
      begin
        myColumns[i] :=   Columns[i].Width;
      end;
    end else
    begin
      for i := 0 to COLUMN_COUNT-1 do   // set default column width
      begin
        myColumns[i] :=   COLUMNSIZE[i];
      end;
    end;

    Clear();
    Columns.Clear();    // Erase existing columns

    for i := 0 to COLUMN_COUNT-1 do
    begin
      Columns.Add;
      Columns[i].Width := myColumns[i];
      Columns[i].AutoSize := false;
      Columns[i].Caption := COLUMN_NAMES[i];
    end;

    for i := 0 to FChartTimeSeries.EventData.LDPEventTable.Count - 1 do
    begin
      eventRecord := (FChartTimeSeries.EventData.LDPEventTable[i] as TTimeSeriesLDPEventRecord); // get the event index record
      listItem := Items.Add;
      listItem.Caption := eventRecord.EventName;
      myDateTime := DateToStr(eventRecord.DateTime,myFormat);  // setting date format
      listItem.SubItems.Add(myDateTime);
      DateTimeToString(myDateTime,TIME_WITH_MS,eventRecord.DateTime);
      listItem.SubItems.Add(myDateTime);
      listItem.SubItems.Add(eventRecord.Description);
      listItem.SubItems.Add(eventRecord.EventCode);
    end;
  end;
end;

//
// EVENT HANDLERS
//

{!******************************************************************************
* MODULE NAME ...  ScrollBarChartTransparencyChange
* SYSTEM NAME ...  AcSELerator
* AUTHOR ........  Kent Myers
*
* FUNCTIONAL DESCRIPTION:  Handle event from transparency scroll bar
*
* INPUT PARAMETERS:
*                    Sender : TObject
*
* OUTPUT PARAMETERS:
*                    None:
* RETURN VALUE:
*                    None:
*
* ERRORS:
*
* APPLICATION NOTES & EXAMPLE:
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2008.11.06  Kent Myers        Original implementation
*******************************************************************************}
procedure TfrmChartTimeSeries.ScrollBarChartTransparencyChange(Sender: TObject);
var
  i : integer;
begin
  FChartTimeSeries.EventData.InputSettings.ChartTransparency := ScrollBarChartTransparency.Position;

  for i := 0 to FChartTimeSeries.ChartVariableCount-1 do        // for each channel
  begin
    case FChartTimeSeries.EventData.InputSettings.ChartStyle of
      sctLine:
      begin
      end;
      sctArea:
      begin
        TAreaSeries(ChartHMIDisplay.Series[i]).Transparency := ChartTimeSeries.EventData.InputSettings.ChartTransparency;  // percent transparency
      end;
    else
      begin
        Assert(False, 'Invalid chart style');
      end;
    end;
  end;
end;

{!******************************************************************************
* MODULE NAME ...  FormCloseQuery
* SYSTEM NAME ...  AcSELerator
* AUTHOR ........  Kent Myers
*
* FUNCTIONAL DESCRIPTION:  Prevent destruction of chart form when form closed: [X] selected
*
* INPUT PARAMETERS:
*                    Sender:  Tobject
* OUTPUT PARAMETERS:
*                    CanClose : Boolean   return false to prevent window destruction
* RETURN VALUE:
*                    None:
*
* ERRORS:
*
* APPLICATION NOTES & EXAMPLE:
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2008.09.24  Kent Myers        Original implementation (obtained from internet)
*******************************************************************************}
procedure TfrmChartTimeSeries.FormCloseQuery(Sender: TObject; var CanClose: Boolean);
begin
  CanClose := not FMenuChart;
  if(FMenuChart) then
  begin
    if(Assigned(ChartMenuHandler)) then  // This is a delegate from FormDB
    begin
      ChartMenuHandler(self);  // delete chart instance from master list
    end;
  end else
  begin
    self.Hide;
  end;
end;
{************************************************************************
* NAME:    FormHide
* FUNCTIONAL DESCRIPTION: Hide the device info form
* INPUT PARAMETERS:       none
* OUTPUT PARAMETERS:      none
* RETURN:                 none
************************************************************************}

procedure TfrmChartTimeSeries.FormHide(Sender: TObject);
begin
  FDeviceInfoDialog.Hide;
end;

{!******************************************************************************
* MODULE NAME ...  ButtonChartResetClick
* SYSTEM NAME ...  AcSELerator
* AUTHOR ........  Kent Myers
*
* FUNCTIONAL DESCRIPTION:  Reset chart axis limits
*
* INPUT PARAMETERS:
*                    Sender : TObject
*
* OUTPUT PARAMETERS:
*                    None:
*
* RETURN VALUE:
*                    None:
*
* ERRORS:
*
* APPLICATION NOTES & EXAMPLE:
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2008.09.24  Kent Myers        Original implementation
*******************************************************************************}
procedure TfrmChartTimeSeries.ButtonChartResetClick(Sender: TObject);
begin
  FChartTimeSeries.AutoScaleAxes();
  ChartHMIDisplay.Update();
end;
{!******************************************************************************
* MODULE NAME ...  ButtonEventUpdateClick
* SYSTEM NAME ...  AcSELerator
* AUTHOR ........  Kent Myers
*
* FUNCTIONAL DESCRIPTION:  Update chart based on new event filter settings
*
* INPUT PARAMETERS:
*                    Sender : TObject
*
* OUTPUT PARAMETERS:
*                    None:
*
* RETURN VALUE:
*                    None:
*
* ERRORS:
*
* APPLICATION NOTES & EXAMPLE:
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2008.09.24  Kent Myers        Original implementation
*******************************************************************************}
procedure TfrmChartTimeSeries.ButtonEventUpdateClick(Sender: TObject);
var status : boolean;
begin

  status := false;
  case FChartType of
    ctSSI:
    begin
      GetChartInputControlSettingsSSI;
      status := FChartTimeSeries.FillEventIndexTable(true);   // generate list before sizing panels to force autosize
    end;

    ctLDP:
    begin
      GetChartInputControlSettingsLDP;
      status := FChartTimeSeries.FillEventIndexTable(true);   // generate list before sizing panels to force autosize
    end;
    else
    begin
      Assert(false,'Invalid chart type');
    end;
  end;

  if(status) then        // there is valid data to chart (may or may not be events)
  begin
    case FChartType of
      ctSSI:
      begin
        FillEventListViewSSI(true);
      end;
      ctLDP:
      begin
        FillEventListViewLDP(true);
      end;
    end;
    UpdateChart(true);
  end;
end;
{************************************************************************
* NAME:       ButtonImportClick
* FUNCTIONAL DESCRIPTION: launch open file dialog to import chart data
* INPUT PARAMETERS: Sender : TObject
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************}
procedure TfrmChartTimeSeries.ButtonImportClick(Sender: TObject);
var
  openDialog: TOpenDialog;
  found : boolean;
begin
  openDialog := TOpenDialog.Create(nil);
  try
    openDialog.InitialDir := GetSpecialFolder(CSIDL_PERSONAL);
    openDialog.DefaultExt := '.csv';
    openDialog.Filter :=
      'Chart File (*.csv)|*.csv|' +
      'SSI Report File (*.ssi)|*.ssi|' +
      'LDP Report File (*.ldp)|*.ldp|' +
      'LDP/SSI Report File (*.txt)|*.txt|' +
      'All Files (*.*)|*.*';
    openDialog.FilterIndex := 1;
    openDialog.Title := 'Import Chart Data';

    openDialog.Options := [ofFileMustExist];
    if openDialog.Execute then
    begin
      found := DisplayChartFile(openDialog.FileName);
      if(found) then
      begin
        Caption := Format('Chart %d - %s',[FChartInstance,
                          ExtractFileName(openDialog.FileName)]);
      end else
      begin
        ClearChartWindow;
        ShowMessage('File does not contain valid data.');
      end;
    end;
  finally
    FreeAndNil(openDialog);
  end;
end;
{************************************************************************
* NAME:       ButtonNotepadClick
* FUNCTIONAL DESCRIPTION: launch Notepad window to import ASCII chart data
* INPUT PARAMETERS: Sender : TObject
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************}
procedure TfrmChartTimeSeries.ButtonNotepadClick(Sender: TObject);
begin
  if(FNotepadForm = nil) then
  begin
    FNotepadForm := TFormNotepad.Create(self);
    FNotePadForm.Caption := Format('Chart %d Notepad',[FChartInstance]);
    FNotepadForm.UpdateChartHandler := UpdateChartHandler;   // delegates
    FNotepadForm.UpdateReportHandler := UpdateReportHandler;
    FNotepadForm.RichEditReport.MaxLength := System.MaxInt-2;
    FNotepadForm.BorderIcons := [biSystemMenu,biMinimize,biMaximize];
  end;
  FNotepadForm.Show;
end;
{!******************************************************************************
* MODULE NAME ...  ButtonEventCaptureClick
* SYSTEM NAME ...  AcSELerator
* AUTHOR ........  Kent Myers
*
* FUNCTIONAL DESCRIPTION:  Capture event summary to clipboard
*
* INPUT PARAMETERS:
*                    Sender : TObject
*
* OUTPUT PARAMETERS:
*                    None:
*
* RETURN VALUE:
*                    None:
*
* ERRORS:
*
* APPLICATION NOTES & EXAMPLE:
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2008.09.24  Kent Myers        Original implementation
*******************************************************************************}
procedure TfrmChartTimeSeries.ButtonEventCaptureClick(Sender: TObject);
var
  i,j : integer;
  listItem : TListItem;
  myText : string;
  myStream : TMemoryStream;

  buffer: THandle;
  bufferPtr: Pointer;
  startPos, bytesToCopy: Integer;  // EndPos,
  P: PChar;

begin
  if(ChartTimeSeries.ChartVariableCount = 0) then exit;   // no data

  myStream := TMemoryStream.Create();
  try
    with ListViewEventList do
    begin
      myText := Format('    %s',[TAB]);
      myStream.Write(myText[1],Length(myText));

      for i := 0 to Columns.Count - 1 do
      begin
        myText := Format('%-11s',[Columns[i].Caption]);
        myStream.Write(myText[1],Length(myText));
        myText := TAB;
        myStream.Write(myText[1],1);
      end;
      myText := CR;
      myStream.Write(myText[1],1);

      // fill list view with contents of event index

      for i := 0 to Items.Count-1 do
      begin
        listItem := Items[i];

        if(listItem.Checked) then
        begin
          myText := Format('[X]%s',[TAB]);
        end else
        begin
          myText := Format('[  ]%s',[TAB]);
        end;
        myStream.Write(myText[1],Length(myText));

        myText := Format('Event #%-5d',[i+1]);
        myStream.Write(myText[1],Length(myText));
        myText := TAB;
        myStream.Write(myText[1],1);
        for j := 0 to Columns.Count - 2 do
        begin
          myText := Format('%-11s',[listItem.SubItems[j]]);
          myStream.Write(myText[1],Length(myText));
          myText := TAB;
          myStream.Write(myText[1],1);
        end;
        myText := CR;
        myStream.Write(myText[1],1);
      end;
    end;

    bytesToCopy := myStream.Position;
    startPos := 0;

    try
      ClipBoard.Open;
      buffer := GlobalAlloc(GMEM_MOVEABLE, BytesToCopy + 1);
      try
        bufferPtr := GlobalLock(buffer);
        try
          myStream.Seek(startPos, soFromBeginning); // reset position to beginning
          myStream.Read(bufferPtr^, bytesToCopy);
          P := bufferPtr;
          P[bytesToCopy] := #0;
          Clipboard.SetAsHandle(CF_TEXT, buffer);
        finally
            GlobalUnlock(buffer);
        end;
      except
        GlobalFree(buffer);
        raise;
      end;
    finally
      Clipboard.Close;
    end;
  finally
    myStream.Free();
  end;
end;
{!******************************************************************************
* MODULE NAME ...  ButtonEventClearClick
* SYSTEM NAME ...  AcSELerator
* AUTHOR ........  Kent Myers
*
* FUNCTIONAL DESCRIPTION:  Disable event display checkboxes
*
* INPUT PARAMETERS:
*                    Sender : TObject
*
* OUTPUT PARAMETERS:
*                    None:
* RETURN VALUE:
*                    None:
*
* ERRORS:
*
* APPLICATION NOTES & EXAMPLE:
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2008.09.24  Kent Myers        Original implementation
*******************************************************************************}
procedure TfrmChartTimeSeries.ButtonEventClearClick(Sender: TObject);
var
  i : integer;
  listItem : TListItem;
  indexRecord : TTimeSeriesIndexRecord;
begin

  if(ListViewEventList.Items.Count = 0) then    // nothing to do
  begin
    exit;
  end;

  for i := 0 to ListViewEventList.Items.Count-1 do
  begin
    listItem := ListViewEventList.Items[i];
    listItem.Checked := false;
  end;
  for i := 0 to FChartTimeSeries.EventData.IndexTable.Count - 1 do
  begin
    indexRecord := (FChartTimeSeries.EventData.IndexTable[i] as TTimeSeriesIndexRecord); // get the event index record
    indexRecord.Checked := false;
  end;
  UpdateChart(true);
end;
{************************************************************************
* NAME:  sbSelectAllClick
* FUNCTIONAL DESCRIPTION:  select all items in event summary
* INPUT PARAMETERS: Sender : TObject
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************}
procedure TfrmChartTimeSeries.sbSelectAllClick(Sender: TObject);
var
  i : integer;
  listItem : TListItem;
  indexRecord : TTimeSeriesIndexRecord;
begin

  if(ListViewEventList.Items.Count = 0) then    // nothing to do
  begin
    exit;
  end;

  for i := 0 to ListViewEventList.Items.Count-1 do
  begin
    listItem := ListViewEventList.Items[i];
    listItem.Checked := true;
  end;
  for i := 0 to FChartTimeSeries.EventData.IndexTable.Count - 1 do
  begin
    indexRecord := (FChartTimeSeries.EventData.IndexTable[i] as TTimeSeriesIndexRecord); // get the event index record
    indexRecord.Checked := true;
  end;
  UpdateChart(true);
end;
{!******************************************************************************
* MODULE NAME ...  WMExitSizeMove
* SYSTEM NAME ...  AcSELerator
* AUTHOR ........  Kent Myers
*
* FUNCTIONAL DESCRIPTION:  Preserve latest window size and position
*
* INPUT PARAMETERS:
*                    Message : TMessage  (unused)
*
* OUTPUT PARAMETERS:
*
* RETURN VALUE:
*                    None:
*
* ERRORS:
*
* APPLICATION NOTES & EXAMPLE:
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2008.12.15  Kent Myers        Original implementation
*******************************************************************************}

procedure TfrmChartTimeSeries.WMExitSizeMove(var Message: TMessage) ;
begin
  Assert(FChartTimeSeries.EventData <> nil,'EventData is nil');
// preserve latest window position
  FChartTimeSeries.EventData.InputSettings.WindowX := Left;
  FChartTimeSeries.EventData.InputSettings.WindowY := Top;
  //   Preserve latest window dimensions
  FChartTimeSeries.EventData.InputSettings.WindowHeight := Height;
  FChartTimeSeries.EventData.InputSettings.WindowWidth := Width;
end; (*WMExitSizeMove*)

{!******************************************************************************
* MODULE NAME ...  FormMouseWheel
* SYSTEM NAME ...  AcSELerator
* AUTHOR ........  Kent Myers
*
* FUNCTIONAL DESCRIPTION:  Event handler for mouse wheel events
*
* INPUT PARAMETERS:
*                    Sender : TObject
*                    Shift : TShiftState   set containing list of depressed control keys
*                    WheelDelta : Integer  change in mouse wheel position
*                    MousePos : TPoint     screen coord. of mouse
*
* OUTPUT PARAMETERS:
*                    Handled: Boolean   true/false indicating whether event was handled
*
* RETURN VALUE:
*                    None:
*
* ERRORS:
*
* APPLICATION NOTES & EXAMPLE:
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2008.09.24  Kent Myers        Original implementation
*******************************************************************************}
procedure TfrmChartTimeSeries.FormMouseWheel(Sender: TObject; Shift: TShiftState;
  WheelDelta: Integer; MousePos: TPoint; var Handled: Boolean);
begin
   ChartHMIDisplayMouseWheel(Sender, Shift, WheelDelta,MousePos,Handled);
end;
{************************************************************************
* NAME:  FormResize
* FUNCTIONAL DESCRIPTION:   test for maximized chart window and adjust location
* INPUT PARAMETERS: Sender : TObject
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************}

procedure TfrmChartTimeSeries.FormResize(Sender: TObject);
begin
  if(self.WindowState = wsMaximized) then
  begin
    self.Top := 0;
    self.Left := 0;
  end;
end;

{!******************************************************************************
* MODULE NAME ...  ChartHMIDisplayMouseWheel
* SYSTEM NAME ...  AcSELerator
* AUTHOR ........  Kent Myers
*
* FUNCTIONAL DESCRIPTION:  Event handler for mouse wheel events
*
* INPUT PARAMETERS:
*                    Sender : TObject
*                    Shift : TShiftState   set containing list of depressed control keys
*                    WheelDelta : Integer  change in mouse wheel position
*                    MousePos : TPoint     screen coord. of mouse
*
* OUTPUT PARAMETERS:
*                    Handled: Boolean   true/false indicating whether event was handled
*
* RETURN VALUE:
*                    None:
*
* ERRORS:
*
* APPLICATION NOTES & EXAMPLE:
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2008.09.24  Kent Myers        Original implementation
*******************************************************************************}
procedure TfrmChartTimeSeries.ChartHMIDisplayMouseWheel(Sender: TObject;
  Shift: TShiftState; WheelDelta: Integer; MousePos: TPoint;
  var Handled: Boolean);
var
  zoom2D : double;
  myWheelDelta : integer;
begin
  if(WindowFromPoint(MousePos) = FChartTimeSeries.ChartDisplay.Handle) then
  begin
    Handled := true;  // event was handled here - don't handle it for other controls
    myWheelDelta := WheelDelta div 100;
    // see if 3D checkbox checked or raw mouse wheel event (no control keys pressed)
    if(not FChartTimeSeries.ChartDisplay.View3D or
      (not (ssShift in Shift) and not (ssCtrl in Shift) and not (ssAlt in Shift))) then  // 2D chart wheel event
    begin
      zoom2D := 100.0 + myWheelDelta;
      FChartTimeSeries.ChartDisplay.ZoomPercent(zoom2D);
    end else
    begin
      if ((ssShift in Shift) and (ssCtrl in Shift)) then  // zoom, X & Y rotate simultaneously
      begin
        FChartTimeSeries.ChartDisplay.View3DOptions.Elevation :=
            FChartTimeSeries.ChartDisplay.View3DOptions.Elevation + myWheelDelta;
        FChartTimeSeries.ChartDisplay.View3DOptions.Rotation :=
            FChartTimeSeries.ChartDisplay.View3DOptions.Rotation + myWheelDelta;
        FChartTimeSeries.ChartDisplay.View3DOptions.Zoom :=
            FChartTimeSeries.ChartDisplay.View3DOptions.Zoom + myWheelDelta;
      end else
      if(ssShift in Shift) then
      begin
        FChartTimeSeries.ChartDisplay.View3DOptions.Zoom :=
            FChartTimeSeries.ChartDisplay.View3DOptions.Zoom + myWheelDelta;
      end else
      if (ssCtrl in Shift) then
      begin
        FChartTimeSeries.ChartDisplay.View3DOptions.Rotation :=
            FChartTimeSeries.ChartDisplay.View3DOptions.Rotation + myWheelDelta;
      end else
      if (ssAlt in Shift) then
      begin
        FChartTimeSeries.ChartDisplay.View3DOptions.Elevation :=
            FChartTimeSeries.ChartDisplay.View3DOptions.Elevation + myWheelDelta;
      end;
    end;
  end;
end;
{!******************************************************************************
* MODULE NAME ...  ChartHMIDisplayMouseDown
* SYSTEM NAME ...  AcSELerator
* AUTHOR ........  Kent Myers
*
* FUNCTIONAL DESCRIPTION:  Event handler for mouse down events
*
* INPUT PARAMETERS:
*                    Sender : TObject
*                    Button : TMouseButton
*                    Shift : TShiftState   set containing list of depressed control keys
*                    X,Y : Integer     screen coord. of mouse
*
* OUTPUT PARAMETERS:
*                    Handled: Boolean   true/false indicating whether event was handled
*
* RETURN VALUE:
*                    None:
*
* ERRORS:
*
* APPLICATION NOTES & EXAMPLE:
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2008.09.24  Kent Myers        Original implementation
*******************************************************************************}
procedure TfrmChartTimeSeries.ChartHMIDisplayMouseDown(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin

  if((Button = mbRight) and ((ssCtrl in Shift) or (ssShift in Shift) or (ssAlt in Shift))) then
  begin
    FChartTimeSeries.ChartDisplay.AllowPanning := pmNone;  // disable during rotation
    FChartTimeSeries.MouseX := X;
    FChartTimeSeries.MouseY := Y;
  end;

  if(Button = mbLeft) then
  begin
    if(FChartTimeSeries.ChartRotate or FChartTimeSeries.ChartTranslate or
       FChartTimeSeries.ChartZoom or FChartTimeSeries.ChartThickness) then
    begin
      FChartTimeSeries.ChartActive := true;
      FChartTimeSeries.MouseX := X;
      FChartTimeSeries.MouseY := Y;
    end;
  end;
end;
{!******************************************************************************
* MODULE NAME ...  ChartHMIDisplayMouseUp
* SYSTEM NAME ...  AcSELerator
* AUTHOR ........  Kent Myers
*
* FUNCTIONAL DESCRIPTION:  Event handler for mouse up events
*
* INPUT PARAMETERS:
*                    Sender : TObject
*                    Button : TMouseButton
*                    Shift : TShiftState   set containing list of depressed control keys
*                    X,Y : Integer     screen coord. of mouse
*
* OUTPUT PARAMETERS:
*                    Handled: Boolean   true/false indicating whether event was handled
*
* RETURN VALUE:
*                    None:
*
* ERRORS:
*
* APPLICATION NOTES & EXAMPLE:
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2008.09.24  Kent Myers        Original implementation
*******************************************************************************}
procedure TfrmChartTimeSeries.ChartHMIDisplayMouseUp(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin

  if(Button = mbRight) then
  begin
    FChartTimeSeries.ChartDisplay.AllowPanning := pmBoth;  // re-enable after rotation
  end;

  if(Button = mbLeft) then
  begin
    FChartTimeSeries.ChartActive := false;  // disable dynamic operations
  end;
end;
{!******************************************************************************
* MODULE NAME ...  ChartHMIDisplayMouseMove
* SYSTEM NAME ...  AcSELerator
* AUTHOR ........  Kent Myers
*
* FUNCTIONAL DESCRIPTION:  Event handler for mouse move events
*
* INPUT PARAMETERS:
*                    Sender : TObject
*                    Shift : TShiftState   set containing list of depressed control keys
*                    X,Y : Integer     screen coord. of mouse
*
* OUTPUT PARAMETERS:
*                    Handled: Boolean   true/false indicating whether event was handled
*
* RETURN VALUE:
*                    None:
*
* ERRORS:
*
* APPLICATION NOTES & EXAMPLE:
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2008.09.24  Kent Myers        Original implementation
* 2008.12.18  Kent Myers        Add extra title for mouse cursor data feedback if necessary
*******************************************************************************}
procedure TfrmChartTimeSeries.ChartHMIDisplayMouseMove(Sender: TObject;
  Shift: TShiftState; X, Y: Integer);
var
  deltaX, deltaY : integer;
  zoom2D : double;
begin
  Assert(FChartTimeSeries.EventData <> nil,'EventData is nil');
  Assert(FChartTimeSeries.EventData.InputSettings <> nil,'InputSettings is nil');

  deltaX := X - FChartTimeSeries.MouseX;
  deltaY := Y - FChartTimeSeries.MouseY;
  FChartTimeSeries.MouseX := X;
  FChartTimeSeries.MouseY := Y;
  if(((ssLeft in Shift) and
       FChartTimeSeries.ChartActive and
       (FChartTimeSeries.ChartRotate or
        FChartTimeSeries.ChartTranslate or
        FChartTimeSeries.ChartZoom or
        FChartTimeSeries.ChartThickness))) then
  begin
    with FChartTimeSeries.ChartDisplay.View3DOptions do
    begin
      if(FChartTimeSeries.EventData.InputSettings.Chart3D and
         FChartTimeSeries.ChartActive) then
      begin
        if(FChartTimeSeries.ChartRotate) then
        begin
          Orthogonal := false;
          Rotation := Rotation + deltaX;
          Elevation := Elevation - deltaY;
        end;
        if(FChartTimeSeries.ChartTranslate) then
        begin
          HorizOffset := HorizOffset + deltaX;
          VertOffset := VertOffset + deltaY;
        end;
        if(FChartTimeSeries.ChartZoom) then
        begin
          Zoom := Zoom + deltaX;
          Zoom := Zoom - deltaY;
        end;
        if(FChartTimeSeries.ChartThickness) then
        begin
          if(((FChartTimeSeries.ChartDisplay.Chart3DPercent + deltaX) < 101) and
             ((FChartTimeSeries.ChartDisplay.Chart3DPercent + deltaX) > 0)) then
          begin
             FChartTimeSeries.ChartDisplay.Chart3DPercent :=
                 FChartTimeSeries.ChartDisplay.Chart3DPercent + deltaX;
          end else
          begin
            if((FChartTimeSeries.ChartDisplay.Chart3DPercent + deltaX) >= 101) then
            begin
              FChartTimeSeries.ChartDisplay.Chart3DPercent := 100;
            end else
            begin
              FChartTimeSeries.ChartDisplay.Chart3DPercent := 1;
            end;
          end;
        end;
      end;
    end;
  end else
  begin
    if((ssRight in Shift) and ((ssCtrl in Shift) or (ssAlt in Shift) or (ssShift in Shift))) then
    begin
      with FChartTimeSeries.ChartDisplay.View3DOptions do
      begin
        if(ssShift in Shift) then     // do 3D rotation
        begin
          Elevation := Elevation - deltaY;
          Rotation := Rotation + deltaX;
        end else
        if(ssCtrl in Shift) then      // do 2D zoom
        begin
          zoom2D := 100.0 - deltaY;
          FChartTimeSeries.ChartDisplay.ZoomPercent(zoom2D);
        end else
        if(ssAlt in Shift) then       // do 3D zoom
        begin
          FChartTimeSeries.ChartDisplay.View3DOptions.Zoom :=
              FChartTimeSeries.ChartDisplay.View3DOptions.Zoom - deltaY;
          FChartTimeSeries.ChartDisplay.View3DOptions.Zoom :=
              FChartTimeSeries.ChartDisplay.View3DOptions.Zoom + deltaX;
        end;
      end;
    end else
    begin   // get X,Y values from curve
      DisplayMouseQueryFeedback;
    end;
  end;
  FChartTimeSeries.EventData.InputSettings.Rotation := FChartTimeSeries.ChartDisplay.View3DOptions.Rotation;
  FChartTimeSeries.EventData.InputSettings.Zoom := FChartTimeSeries.ChartDisplay.View3DOptions.Zoom;
  FChartTimeSeries.EventData.InputSettings.HorizOffset := FChartTimeSeries.ChartDisplay.View3DOptions.HorizOffset;
  FChartTimeSeries.EventData.InputSettings.VertOffset := FChartTimeSeries.ChartDisplay.View3DOptions.VertOffset;
  FChartTimeSeries.EventData.InputSettings.Elevation := FChartTimeSeries.ChartDisplay.View3DOptions.Elevation;
  FChartTimeSeries.EventData.InputSettings.Percent3D := FChartTimeSeries.ChartDisplay.Chart3DPercent;
end;
{************************************************************************
* NAME:   DisplayMouseQueryFeedback
* FUNCTIONAL DESCRIPTION:  Display feedback for chart object at mouse cursor location
* INPUT PARAMETERS: none
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************}
procedure TfrmChartTimeSeries.DisplayMouseQueryFeedback;
var
  P:TPoint;
  found, isCBEMA : boolean;
  myX, myY : double;
  i, j, ecount, count, nActiveRecords, match, index, lastRecord : integer;
  indexRecord : TTimeSeriesIndexRecord;
  dataRecord : TTimeSeriesRecord;
  myLabel, myDT : string;
  myFormat : TFormatSettings;
begin
  FChartTimeSeries.EventData.InitDateTimeFormat(ftInternational,myFormat);

  isCBEMA := false;
  case FChartType of
    ctUndefined, ctLDP:
    begin
    end;
    ctSSI:
    begin
      if(TTimeSeriesInputSettingsSSI(FChartTimeSeries.EventData.InputSettings).ChartCBEMA) then
      begin
        isCBEMA := true;
      end;
    end;
    else
    begin
      Assert(false,'Invalid chart type');
    end;
  end;

  {NOTE: A semi-colon separator is used (instead of a comma) when feeding back
   a list of data values in order to avoid confusion when the native language
   uses a comma instead of a decimal point}

  if(FChartTimeSeries.EventData.InputSettings.ChartQuery) then
  begin
    if(FChartTimeSeries.ChartDisplay.Title.Text.Count <= 1) then
    begin
      FChartTimeSeries.ChartDisplay.Title.Text.Add(' ');   // add title for feedback
    end;
    P:=FChartTimeSeries.ChartDisplay.GetCursorPos;

    ecount := GetActiveEventCount;   // number of active events

    // Compare mouse location to each point in each series

    for i := 0 to FChartTimeSeries.ChartDisplay.SeriesCount - 1 do
    begin
      match := FChartTimeSeries.ChartDisplay.Series[i].Clicked(P.X,P.Y);
      if match<>-1 then     // match is matching record in series
      begin
        FChartTimeSeries.CursorTool.Series := FChartTimeSeries.ChartDisplay.Series[i];
        if((not isCBEMA) and
            not (FChartTimeSeries.ChartDisplay.Series[i] is TBarSeries)) then
        begin
          match := FChartTimeSeries.CursorTool.SnapToPoint();   // not a point series, must identify data point
        end;
        myX := FChartTimeSeries.ChartDisplay.Series[i].XValue[match];
        myY := FChartTimeSeries.ChartDisplay.Series[i].YValue[match];

        if((FChartType = ctLDP) and (ChartTimeSeries.ChartDisplay.Series[i] is TPointSeries)) then
        begin
          index := FChartTimeSeries.GetEventRecordFromRecord(myX);
          if(index >= 0) then
          begin
            indexRecord := (FChartTimeSeries.EventData.IndexTable[index] as TTimeSeriesIndexRecord);
            myDT := DateToStr(indexRecord.DateTime,myFormat) + ' ' + FormatDateTime(TIME_WITH_MS,indexRecord.DateTime);
            myLabel := Format('Event %s (%s) = %s',
                        [indexRecord.EventName,myDT,indexRecord.Description]);
          end;
        end else
        if(not isCBEMA and FChartTimeSeries.EventData.InputSettings.ChartAxisDateTime) then    // display date/time X axis labels
        begin      // Indep. Var. is Date/Time
          // Determine which record this data point refers to
          found := false;
          if(ChartType = ctSSI) then // update match based on event records
          begin
            nActiveRecords := 0;
            lastRecord := 0;   // upper limit of last active parent record
            for j := 0 to FChartTimeSeries.EventData.IndexTable.Count - 1 do
            begin
              indexRecord := (FChartTimeSeries.EventData.IndexTable[j] as TTimeSeriesIndexRecord);  // get event summary record
              if((ecount > 0) and not indexRecord.Checked) then continue;  // skip inactive record
              if(indexRecord.LastRecord <= lastRecord) then continue;  // skip sub-event
              lastRecord := indexRecord.LastRecord; // remember last record of active parent event
              count := indexRecord.LastRecord - indexRecord.FirstRecord + 1;  // number of records in this event
              if(not found and (match > nActiveRecords) and
                (match <= (nActiveRecords + count))) then  // record is in this interval
              begin
                found := true;      // record is in this event interval
                match := indexRecord.FirstRecord + match - nActiveRecords -1;
                break;
              end;
              Inc(nActiveRecords,count);  // number of active records
            end;
          end else
          begin
            found := true;  // LDP just use match already found
          end;

          if(found) then
          begin
            dataRecord := (FChartTimeSeries.EventData.RawDataTable[match] as TTimeSeriesRecord);
            if(FChartType = ctSSI) then
            begin
              myDT := DateToStr(dataRecord.DateTime,myFormat) + ' ' + FormatDateTime(TIME_WITH_MS,dataRecord.DateTime);
            end else
            begin
              myDT := DateToStr(dataRecord.DateTime,myFormat) + ' ' + FormatDateTime(TIME_WITH_SEC,dataRecord.DateTime);
            end;
            myLabel := Format('%s=(%s; %.2f)',[FChartTimeSeries.ChartDisplay.Series[i].Title,myDT,myY]);
          end;
        end else
        begin        // Indep. var. is record number or duration (if CBEMA)
          if(ChartType = ctLDP) then     // display LDP channel info
          begin
              myLabel := Format('%s=(%.0f; %.2f)',[FChartTimeSeries.ChartDisplay.Series[i].Title,myX,myY]);
          end else
          begin
            if(isCBEMA) then     // display CBEMA event marker info
            begin
              if(i = 0) then    // first series is the event marker
              begin
                myLabel := Format('%s %s=(Duration=%.3f; Magnitude=%.1f)',
                [FChartTimeSeries.ChartDisplay.Series[i].Title,
                ChartTimeSeries.ChartDisplay.Series[i].Marks[match].Text[0],myX,myY]);
              end else           // display sag/swell limit info
              begin
                myLabel := Format('%s =(Duration=%.3f; Magnitude=%.1f)',
                [FChartTimeSeries.ChartDisplay.Series[i].Title,myX,myY]);
              end;
            end else             // display SSI channel series info
            begin
              myLabel := Format('%s=(%.0f; %.2f)',   // Rec #, Magnitude
              [FChartTimeSeries.ChartDisplay.Series[i].Title,myX,myY]);
            end;
          end;
        end;

        FChartTimeSeries.ChartDisplay.Title.Text[FChartTimeSeries.ChartDisplay.Title.Text.Count-1] := myLabel;
      end;
    end;
  end;
end;
{!******************************************************************************
* MODULE NAME ...  SpeedButtonPointerClick
* SYSTEM NAME ...  AcSELerator
* AUTHOR ........  Kent Myers
*
* FUNCTIONAL DESCRIPTION:  Event handler for pointer button in chart toolbar
*
* INPUT PARAMETERS:
*                    Sender : TObject
*
* OUTPUT PARAMETERS:
*                    None:
* RETURN VALUE:
*                    None:
*
* ERRORS:
*
* APPLICATION NOTES & EXAMPLE:
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2008.11.12  Kent Myers        Original implementation
*******************************************************************************}
procedure TfrmChartTimeSeries.SpeedButtonPointerClick(Sender: TObject);
const
  TOOLBAR_MSG = '   Drag left button (down/right) to Zoom, Right button to Translate';
begin
  FChartTimeSeries.ChartDisplay.Aspect.HorizOffset := 0;
  FChartTimeSeries.ChartDisplay.Aspect.VertOffset := 0;
  FChartTimeSeries.ChartDisplay.Aspect.Zoom := 100;

  LabelChartToolbar.Caption := TOOLBAR_MSG;
  FChartTimeSeries.ChartRotate := false;
  FChartTimeSeries.ChartTranslate := false;
  FChartTimeSeries.ChartZoom := false;
  FChartTimeSeries.ChartThickness := false;
  FChartTimeSeries.ChartActive := false;
  FChartTimeSeries.ChartDisplay.AllowPanning := pmBoth;  // re-enable panning
  FChartTimeSeries.ChartDisplay.AllowZoom := true;
end;
{!******************************************************************************
* MODULE NAME ...  SpeedButtonTranslateClick
* SYSTEM NAME ...  AcSELerator
* AUTHOR ........  Kent Myers
*
* FUNCTIONAL DESCRIPTION:  Event handler for translate button in chart toolbar
*
* INPUT PARAMETERS:
*                    Sender : TObject
*
* OUTPUT PARAMETERS:
*                    None:
* RETURN VALUE:
*                    None:
*
* ERRORS:
*
* APPLICATION NOTES & EXAMPLE:
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2008.11.12  Kent Myers        Original implementation
*******************************************************************************}
procedure TfrmChartTimeSeries.SpeedButtonTranslateClick(Sender: TObject);
const
  TRANSLATE_MSG = '   Drag Chart to Translate';
begin
  Assert(FChartTimeSeries.EventData <> nil,'EventData is nil');

  LabelChartToolbar.Caption := TRANSLATE_MSG;
  FChartTimeSeries.ChartRotate := false;
  FChartTimeSeries.ChartTranslate := true;
  FChartTimeSeries.ChartZoom := false;
  FChartTimeSeries.ChartThickness := false;
  FChartTimeSeries.ChartDisplay.AllowPanning := pmBoth;  // disable panning
  FChartTimeSeries.ChartDisplay.AllowZoom := false;
  if(not FChartTimeSeries.EventData.InputSettings.Chart3D) then
  begin
    SpeedButton3DClick(Sender);
  end;
end;

{!******************************************************************************
* MODULE NAME ...  SpeedButtonRotateClick
* SYSTEM NAME ...  AcSELerator
* AUTHOR ........  Kent Myers
*
* FUNCTIONAL DESCRIPTION:  Event handler for rotate button in chart toolbar
*
* INPUT PARAMETERS:
*                    Sender : TObject
*
* OUTPUT PARAMETERS:
*                    None:
* RETURN VALUE:
*                    None:
*
* ERRORS:
*
* APPLICATION NOTES & EXAMPLE:
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2008.11.12  Kent Myers        Original implementation
*******************************************************************************}
procedure TfrmChartTimeSeries.SpeedButtonRotateClick(Sender: TObject);
const
  ROTATE_MSG = '   Drag Chart to Rotate';
begin
  Assert(FChartTimeSeries.EventData <> nil,'EventData is nil');

  LabelChartToolbar.Caption := ROTATE_MSG;
  FChartTimeSeries.ChartRotate := true;
  FChartTimeSeries.ChartTranslate := false;
  FChartTimeSeries.ChartZoom := false;
  FChartTimeSeries.ChartThickness := false;
  FChartTimeSeries.ChartDisplay.AllowZoom := false;
  if(not FChartTimeSeries.EventData.InputSettings.Chart3D) then
  begin
    SpeedButton3DClick(Sender);
  end;
end;
{!******************************************************************************
* MODULE NAME ...  SpeedButtonZoomClick
* SYSTEM NAME ...  AcSELerator
* AUTHOR ........  Kent Myers
*
* FUNCTIONAL DESCRIPTION:  Event handler for zoom button in chart toolbar
*
* INPUT PARAMETERS:
*                    Sender : TObject
*
* OUTPUT PARAMETERS:
*                    None:
* RETURN VALUE:
*                    None:
*
* ERRORS:
*
* APPLICATION NOTES & EXAMPLE:
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2008.11.12  Kent Myers        Original implementation
*******************************************************************************}
procedure TfrmChartTimeSeries.SpeedButtonZoomClick(Sender: TObject);
const
  ZOOM_MSG =  '   Drag Chart to Zoom';
begin
  Assert(FChartTimeSeries.EventData <> nil,'EventData is nil');

  LabelChartToolbar.Caption := ZOOM_MSG;
  FChartTimeSeries.ChartRotate := false;
  FChartTimeSeries.ChartTranslate := false;
  FChartTimeSeries.ChartZoom := true;
  FChartTimeSeries.ChartThickness := false;
  FChartTimeSeries.ChartDisplay.AllowZoom := false;
  if(not FChartTimeSeries.EventData.InputSettings.Chart3D) then
  begin
    SpeedButton3DClick(Sender);
  end;
end;
{!******************************************************************************
* MODULE NAME ...  SpeedButtonThicknessClick
* SYSTEM NAME ...  AcSELerator
* AUTHOR ........  Kent Myers
*
* FUNCTIONAL DESCRIPTION:  Event handler for thickness button in chart toolbar
*
* INPUT PARAMETERS:
*                    Sender : TObject
*
* OUTPUT PARAMETERS:
*                    None:
* RETURN VALUE:
*                    None:
*
* ERRORS:
*
* APPLICATION NOTES & EXAMPLE:
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2008.11.12  Kent Myers        Original implementation
*******************************************************************************}
procedure TfrmChartTimeSeries.SpeedButtonThicknessClick(Sender: TObject);
const
  DEPTH_MSG = '   Drag Chart to change Depth';
begin
  Assert(FChartTimeSeries.EventData <> nil,'EventData is nil');

  LabelChartToolbar.Caption := DEPTH_MSG;
  FChartTimeSeries.ChartRotate := false;
  FChartTimeSeries.ChartTranslate := false;
  FChartTimeSeries.ChartZoom := false;
  FChartTimeSeries.ChartThickness := true;
  FChartTimeSeries.ChartDisplay.AllowZoom := false;
  if(not FChartTimeSeries.EventData.InputSettings.Chart3D) then
  begin
    SpeedButton3DClick(Sender);
  end;
end;
{!******************************************************************************
* MODULE NAME ...  SpeedButtonCopyClick
* SYSTEM NAME ...  AcSELerator
* AUTHOR ........  Kent Myers
*
* FUNCTIONAL DESCRIPTION:  Event handler for copy button in chart toolbar
*
* INPUT PARAMETERS:
*                    Sender : TObject
*
* OUTPUT PARAMETERS:
*                    None:
* RETURN VALUE:
*                    None:
*
* ERRORS:
*
* APPLICATION NOTES & EXAMPLE:
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2008.11.12  Kent Myers        Original implementation
*******************************************************************************}
procedure TfrmChartTimeSeries.SpeedButtonCopyClick(Sender: TObject);
begin
  FChartTimeSeries.ChartDisplay.BufferedDisplay := False;
  FChartTimeSeries.ChartDisplay.CopyToClipboardBitmap;
  FChartTimeSeries.ChartDisplay.BufferedDisplay := True;
end;
{!******************************************************************************
* MODULE NAME ...  SpeedButtonCursorClick
* SYSTEM NAME ...  AcSELerator
* AUTHOR ........  Kent Myers
*
* FUNCTIONAL DESCRIPTION:  event handler for cursor query
*
* INPUT PARAMETERS:
*                    Sender : TObject
*
* OUTPUT PARAMETERS:
*                    None:
* RETURN VALUE:
*                    None:
*
* ERRORS:
*
* APPLICATION NOTES & EXAMPLE:
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2008.11.12  Kent Myers        Original implementation
*******************************************************************************}
procedure TfrmChartTimeSeries.SpeedButtonCursorClick(Sender: TObject);
begin
  Assert(FChartTimeSeries.EventData <> nil,'EventData is nil');

  FChartTimeSeries.EventData.InputSettings.ChartQuery := not FChartTimeSeries.EventData.InputSettings.ChartQuery;
  FChartTimeSeries.CursorTool.Active := FChartTimeSeries.EventData.InputSettings.ChartQuery;
  FChartTimeSeries.CursorTool.Snap := FChartTimeSeries.EventData.InputSettings.ChartQuery;

  if(FChartTimeSeries.EventData.InputSettings.ChartQuery) then
  begin
    FChartTimeSeries.ChartDisplay.Title.Text.Add(' ');   // add title for feedback
  end else
  begin
    FChartTimeSeries.InitChartTitles;      // reset titles to original
  end;
end;
{!******************************************************************************
* MODULE NAME ...  SpeedButtonDateTimeClick
* SYSTEM NAME ...  AcSELerator
* AUTHOR ........  Kent Myers
*
* FUNCTIONAL DESCRIPTION:  event handler for date/time labels
*
* INPUT PARAMETERS:
*                    Sender : TObject
*
* OUTPUT PARAMETERS:
*                    None:
* RETURN VALUE:
*                    None:
*
* ERRORS:
*
* APPLICATION NOTES & EXAMPLE:
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2008.11.12  Kent Myers        Original implementation
*******************************************************************************}
procedure TfrmChartTimeSeries.SpeedButtonDateTimeClick(Sender: TObject);
begin
  Assert(FChartTimeSeries.EventData <> nil,'EventData is nil');

  FChartTimeSeries.EventData.InputSettings.ChartAxisDateTime := not FChartTimeSeries.EventData.InputSettings.ChartAxisDateTime;
  SetDateTimeLabels();
end;
{!******************************************************************************
* MODULE NAME ...  SpeedButton3DClick
* SYSTEM NAME ...  AcSELerator
* AUTHOR ........  Kent Myers
*
* FUNCTIONAL DESCRIPTION:  Event handler for 3D button in chart toolbar
*
* INPUT PARAMETERS:
*                    Sender : TObject
*
* OUTPUT PARAMETERS:
*                    None:
* RETURN VALUE:
*                    None:
*
* ERRORS:
*
* APPLICATION NOTES & EXAMPLE:
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2008.11.12  Kent Myers        Original implementation
*******************************************************************************}
procedure TfrmChartTimeSeries.SpeedButton3DClick(Sender: TObject);
begin
  Assert(FChartTimeSeries.EventData <> nil,'EventData is nil');

  FChartTimeSeries.EventData.InputSettings.Chart3D := not FChartTimeSeries.EventData.InputSettings.Chart3D;
  FChartTimeSeries.ChartDisplay.View3D := FChartTimeSeries.EventData.InputSettings.Chart3D;
  SetChartLineStyles();
end;
{!******************************************************************************
* MODULE NAME ...  SpeedButtonEditClick
* SYSTEM NAME ...  AcSELerator
* AUTHOR ........  Kent Myers
*
* FUNCTIONAL DESCRIPTION:  Event handler for edit button in chart toolbar
*
* INPUT PARAMETERS:
*                    Sender : TObject
*
* OUTPUT PARAMETERS:
*                    None:
* RETURN VALUE:
*                    None:
*
* ERRORS:
*
* APPLICATION NOTES & EXAMPLE:
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2008.11.12  Kent Myers        Original implementation
*******************************************************************************}
procedure TfrmChartTimeSeries.SpeedButtonEditClick(Sender: TObject);
begin
  EditChartPage(Self, FChartTimeSeries.ChartDisplay, teeEditPagingPage);  // display chart edit dialog
end;
{************************************************************************
* NAME:  SpeedButtonInformationClick
* FUNCTIONAL DESCRIPTION: Display device Information
* INPUT PARAMETERS: none
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************}
procedure TfrmChartTimeSeries.SpeedButtonInformationClick(Sender: TObject);
begin
  if(FDeviceInfoDialog.TabControlDeviceInfo.Tabs.Count > 0) then
  begin
    if(not FDeviceInfoDialog.Visible) then
    begin
      ResetDeviceInformation;
      FDeviceInfoDialog.Show;
    end else
    begin
      FDeviceInfoDialog.Hide;
    end;
  end;
end;
{************************************************************************
* NAME:  ResetDeviceInformation
* FUNCTIONAL DESCRIPTION: Reset device information dialog content
* INPUT PARAMETERS: none
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************}
procedure TfrmChartTimeSeries.ResetDeviceInformation;
begin
  ShowDeviceInformation;
  if(ChartType = ctLDP) then
  begin
    ShowChannelInformation;
    ShowEventInformation;
  end;
  FDeviceInfoDialog.ResetDeviceInfoTab;
end;

{************************************************************************
* NAME:  SpeedButtonLabelClick
* FUNCTIONAL DESCRIPTION: Enable/Disable event/CBEMA labels
* INPUT PARAMETERS: none
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************}
procedure TfrmChartTimeSeries.SpeedButtonLabelClick(Sender: TObject);
var
  BarSeries : TBarSeries;
  PointSeries : TPointSeries;
begin
  Assert(FChartTimeSeries.EventData <> nil,'EventData is nil');
  if(FChartTimeSeries.EventData.RawDataTable.Count < 1) then  // no data loaded into chart
  begin
    exit;
  end;

  FChartTimeSeries.EventData.InputSettings.ChartLabels :=
      not ChartTimeSeries.EventData.InputSettings.ChartLabels;

  if(FChartType = ctSSI) then  // SSI has a bar series with event indicators
  begin
    if(TTimeSeriesInputSettingsSSI(FChartTimeSeries.EventData.InputSettings).ChartCBEMA) then
    begin
      Assert((ChartHMICBEMA.Series[0] is TPointSeries),'Not a Point Series');

      PointSeries := TPointSeries (ChartHMICBEMA.Series[0]);
      PointSeries.Marks.Visible := FChartTimeSeries.EventData.InputSettings.ChartLabels;
    end else
    begin
      if(ChartHMIDisplay.Series[ChartHMIDisplay.SeriesCount-1] is TBarSeries) then
      begin
        BarSeries := TBarSeries (ChartHMIDisplay.Series[ChartHMIDisplay.SeriesCount-1]);
        BarSeries.Marks.Visible := FChartTimeSeries.EventData.InputSettings.ChartLabels;
      end;
    end;
  end;
  if((FChartType = ctLDP) and
     (ChartHMIDisplay.Series[ChartHMIDisplay.SeriesCount-1] is TPointSeries)) then
  begin
    PointSeries := TPointSeries (ChartHMIDisplay.Series[ChartHMIDisplay.SeriesCount-1]);
    PointSeries.Marks.Visible := FChartTimeSeries.EventData.InputSettings.ChartLabels;
  end;
end;

{!******************************************************************************
* MODULE NAME ...  SpeedButtonResetClick
* SYSTEM NAME ...  AcSELerator
* AUTHOR ........  Kent Myers
*
* FUNCTIONAL DESCRIPTION:  Event handler for reset button in chart toolbar
*
* INPUT PARAMETERS:
*                    Sender : TObject
*
* OUTPUT PARAMETERS:
*                    None:
* RETURN VALUE:
*                    None:
*
* ERRORS:
*
* APPLICATION NOTES & EXAMPLE:
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2008.11.12  Kent Myers        Original implementation
*******************************************************************************}
procedure TfrmChartTimeSeries.SpeedButtonResetClick(Sender: TObject);
begin
  Assert(FChartTimeSeries.EventData <> nil,'EventData is nil');

  FChartTimeSeries.AutoScaleAxes();
  with FChartTimeSeries.ChartDisplay do
  begin
    Aspect.HorizOffset := 0;
    Aspect.VertOffset := 0;
    Aspect.Zoom := 100;
    Chart3DPercent := 10;
  end;
  with FChartTimeSeries.ChartDisplay.View3DOptions do
  begin
    Orthogonal := false;
    Rotation := 0;
    Elevation := 0;
  end;
  FChartTimeSeries.ChartDisplay.Update();
end;
{************************************************************************
* NAME:   ChartHMIDisplayGetAxisLabel
* FUNCTIONAL DESCRIPTION:  This function is used to revise the x-axis label
*                          if the Date/Time option is active.
* INPUT PARAMETERS:
*                    Sender : TChartAxis
*                    Series : TChartSeries
*                    ValueIndex : Integer
*                    LabelText  : string
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************}
procedure TfrmChartTimeSeries.ChartHMIDisplayGetAxisLabel(Sender: TChartAxis;
  Series: TChartSeries; ValueIndex: Integer; var LabelText: string);
var
  value, delta : double;
  ivalue,idelta,jDate,jDate1,jdate2,lastRecord : integer;
  dataRecord : TTimeSeriesRecord;
  indexRecord : TTimeSeriesIndexRecord;
  date_label, no_label, found, minFound, maxFound : boolean;
  i,i1,i2,number,count,ecount,ecount2,Left,Right,iMin,iMax : integer;
  eLimits : array of TLimits;     // lower/upper limit of each event interval
  myLabel,TimeString : string;
  allEvents : boolean;
  myFormat : TFormatSettings;
begin
  FChartTimeSeries.EventData.InitDateTimeFormat(ftInternational,myFormat);
  DateFormat2DigitYear(myFormat);
{
*      Records are stored and sorted by record number, so the
*      record number must be used to look up each record,
*      determine the time/date and produce a label for the
*      x-axis.  Special logic is used to determine whether
*      two consecutive labels (the current label and the label to the left)
*      have the same date.  If so, the current label will not include the date.
}

  Assert(FChartTimeSeries.EventData <> nil,'EventData is nil');

  if(FChartTimeSeries.EventData.InputSettings.ChartAxisDateTime and
    (Sender = FChartTimeSeries.ChartDisplay.BottomAxis)) then
  begin
    case FChartType of
      ctUndefined, ctLDP:
      begin

      end;
      ctSSI:
      begin
        if(TTimeSeriesInputSettingsSSI(FChartTimeSeries.EventData.InputSettings).ChartCBEMA) then
        begin
          exit; // just use default labels for CBEMA chart
        end;
      end;
      else
      begin
        Assert(false,'Invalid chart type');
      end;
    end;

    myLabel := AnsiReplaceStr(LabelText,',','');  // remove commas form present label (thousands)
    value := NonLocalStrToFloat(myLabel);
    ivalue := Round(value);         // record number in range 0 to total # active records
    allEvents := true;   // assume all events are displayed
    iMin := Trunc(FChartTimeSeries.ChartDisplay.BottomAxis.Minimum);  // minimum record number on X axis
    iMax := Trunc(FChartTimeSeries.ChartDisplay.BottomAxis.Maximum);  // maximum record number on X axis

    // update present record limits to actual record limits based on active events

    number := 0;  // number of active event records
    ecount := 0;  // number of active events
    Left := 1;   // lower limit if no active events
    Right := FChartTimeSeries.EventData.RawDataTable.Count;  // upper limit if no active events
    if((FChartType = ctSSI) and (FChartTimeSeries.EventData.indexTable.Count > 0)) then     // events are present
    begin
      Left := -1;
      lastRecord := 0;
      for i := 0 to FChartTimeSeries.EventData.IndexTable.Count - 1 do
      begin
        indexRecord := (FChartTimeSeries.EventData.IndexTable[i] as TTimeSeriesIndexRecord);  // get event summary record
        if(not indexRecord.Checked) then continue; // event not active, skip it

        inc(ecount);     // increment active event count
        if(Left < 0) then    // Left not found yet, this is start of first active record
        begin
          Left := indexRecord.FirstRecord;
        end;
        if(Right < indexRecord.LastRecord + 1) then   // skip if sub-event has a smaller right bound
        begin
          Right := indexRecord.LastRecord + 1;
        end;
        if(indexRecord.LastRecord > lastRecord) then  // test for sub-event
        begin
          number := number + indexRecord.LastRecord - indexRecord.FirstRecord + 1;   // number of active records so far
        end;
        lastRecord := indexRecord.LastRecord;
      end;
      if(Left < 0) then Left := 0;  // no active events found

      // Present label is outside valid range of record numbers, just return empty label
      if((ivalue < 0) or ((number > 0) and (ivalue > number))) then
      begin
        LabelText := ' ';
        exit;     // record is invalid
      end;

      // find record 'ivalue' in list of active records

      if(ecount > 0) then    // subset of events are active
      begin
        allEvents := false;
        SetLength(eLimits,ecount);
      end else               // all events are active by default
      begin
        SetLength(eLimits,FChartTimeSeries.EventData.indexTable.Count);
      end;

      number := 0;    // initialize number of active records
      ecount2 := 0;   // initialize active event count
      found := false;  // assume value is not in active interval
      minFound := false;  // assume minimum rec # not found yet
      maxFound := false;  // assume maximum rec # not found yet

      // Convert present record number and min/max record numbers to actual record numbers
      lastRecord := 0;  // test for sub-events
      for i := 0 to FChartTimeSeries.EventData.IndexTable.Count - 1 do
      begin
        indexRecord := (ChartTimeSeries.EventData.IndexTable[i] as TTimeSeriesIndexRecord);  // get event summary record
        if(not allEvents and not indexRecord.Checked) then continue;  // skip inactive record

        if(indexRecord.LastRecord <= lastRecord) then continue;  // skip sub-events
        lastRecord := indexRecord.LastRecord;

        eLimits[ecount2].limit1 := indexRecord.FirstRecord;  // remember upper limit of each event interval
        eLimits[ecount2].limit2 := indexRecord.LastRecord;  // remember upper limit of each event interval
        inc(ecount2);     // increment active event count
        count := indexRecord.LastRecord - indexRecord.FirstRecord + 1;  // number of records in this event
        if(not found and (ivalue > number) and (ivalue <= (number + count))) then  // record is in this interval
        begin
          found := true;      // record is in this event interval
          ivalue := indexRecord.FirstRecord + ivalue - number;
        end;
        if(not minFound and (iMin > number) and (iMin <= (number + count))) then
        begin
          minFound := true;     // minimum is in this event interval
          iMin := indexRecord.FirstRecord + iMin - number;
        end;
        if(not maxFound and (iMax > number) and (iMax <= (number + count))) then
        begin
          maxFound := true;     // maximum is in this event interval
          iMax := indexRecord.FirstRecord + iMax - number;
        end;
        number := number + count;  // number of active records
      end;

      if(not minFound) then  // before beginning of the first event
      begin
        iMin := Left;
      end;
      if(not maxFound) then  // after end of the last event
      begin
        iMax := Right;
      end;
    end else
    begin           // no events, all records in use
      found := true;  // assume value is in active interval
      SetLength(eLimits,1);
      eLimits[0].limit1 := 1;
      eLimits[0].limit2 := FChartTimeSeries.EventData.RawDataTable.Count;
    end;

    date_label := false;   // assume we are not displaying the date
    no_label := true;      // assume that no label will be displayed
    if(found and (ivalue >= Left) and (ivalue <= Right)) then
    begin
      no_label := false;  // some label will be displayed
      delta := FChartTimeSeries.ChartXDelta;  // delta obtained in separate event handler to prevent recursion
      idelta := Round(delta)+1;  // record interval between labels
      if(idelta < 1) then
      begin
        idelta := 1;
      end;

      // See if this is the leftmost or rightmost axis label
      if(((ivalue - idelta) <= Left) or ((ivalue + idelta) >= Right) or
         ((ivalue - idelta) <= iMin) or ((ivalue + idelta) >= iMax)) then
      begin
         date_label := true;     // display the date
      end else
      begin
        i1 := FChartTimeSeries.WhichInterval(ivalue - idelta, eLimits);  // event interval containing previous label
        i2 := FChartTimeSeries.WhichInterval(ivalue + idelta, eLimits);  // event interval containing next label

        if(i1 <> i2) then // neighboring labels are from a different event
        begin
          date_label := true;   // display the date at interval boundaries
        end else
        begin
          dataRecord := (FChartTimeSeries.EventData.RawDataTable[ivalue-1] as TTimeSeriesRecord);   // record for current label
          jDate := DateToJulian(dataRecord.DateTime);
          dataRecord := (FChartTimeSeries.EventData.RawDataTable[ivalue-idelta-1] as TTimeSeriesRecord);  // record for previous label
          jDate1 := DateToJulian(dataRecord.DateTime);
          if(jDate <> jDate1) then  // date is different for neighboring labels
          begin
            date_label := true;     // display the date
          end;
        end;
      end;
    end;

    if(no_label) then    // no label should be displayed
    begin
      LabelText := ' ';
    end else
    begin                // some label should be displayed
      dataRecord := (FChartTimeSeries.EventData.RawDataTable[ivalue-1] as TTimeSeriesRecord);  // record for previous label
      if(FChartType = ctSSI) then
      begin
        TimeString := ' ' + FormatDateTime(TIME_WITH_MS,Frac(dataRecord.DateTime));
      end else
      begin
        TimeString := ' ' + FormatDateTime(TIME_WITH_SEC,Frac(dataRecord.DateTime));
      end;
      if(date_label) then   // date should be displayed
      begin
        LabelText := DateToStr(dataRecord.DateTime,myFormat) + TimeString;  // native language
      end else              // only time should be displayed
      begin
        LabelText := TimeString;
      end;
    end;
  end;
end;
{!******************************************************************************
* MODULE NAME ...  ChartHMIDisplayGetNextAxisLabel
* SYSTEM NAME ...  AcSELerator
* AUTHOR ........  Kent Myers
*
* FUNCTIONAL DESCRIPTION:  called just before displaying axis labels
*
* INPUT PARAMETERS:
*                    Sender : TChartAxis
*                    LabelIndex : Integer  index of current label
*                    LabelValue : double   value of current label
*                    Stop  : boolean       true/false to stop callback
*
* OUTPUT PARAMETERS:
*                    None:
* RETURN VALUE:
*                    None:
*
* ERRORS:
*
* APPLICATION NOTES & EXAMPLE:
*  NOTE: This callback is necessary in order to obtain the increment between
*  consecutive labels.  If CalcIncrement() is called in the
*  ChartHMIDisplayGetAxisLabel callback, a recursion occurs during a zoom.
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2008.09.24  Kent Myers        Original implementation
*******************************************************************************}
procedure TfrmChartTimeSeries.ChartHMIDisplayGetNextAxisLabel(Sender: TChartAxis;
  LabelIndex: Integer; var LabelValue: Double; var Stop: Boolean);
begin
  if(Sender = ChartHMIDisplay.BottomAxis) then
  begin
    FChartTimeSeries.ChartXDelta := Sender.CalcIncrement;
  end;
end;
{************************************************************************
* NAME:   ChartHMIDisplayClickLegend
* FUNCTIONAL DESCRIPTION:  update yvalue of SER event markers
* INPUT PARAMETERS: not used
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************}

procedure TfrmChartTimeSeries.ChartHMIDisplayClickLegend(Sender: TCustomChart;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  FChartTimeSeries.UpdateEventScale;
end;

{!******************************************************************************
* MODULE NAME ...  CheckBoxEventBeginClick
* SYSTEM NAME ...  AcSELerator
* AUTHOR ........  Kent Myers
*
* FUNCTIONAL DESCRIPTION:  Enable/disable Begin DateTime controls
*
* INPUT PARAMETERS:
*                    Sender : TObject
*
* OUTPUT PARAMETERS:
*                    None:
* RETURN VALUE:
*                    None:
*
* ERRORS:
*
* APPLICATION NOTES & EXAMPLE:
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2008.09.24  Kent Myers        Original implementation
*******************************************************************************}
procedure TfrmChartTimeSeries.CheckBoxEventBeginClick(Sender: TObject);
begin
  DateTimePickerEventBeginDate.Enabled := CheckBoxEventBegin.Checked;
  DateTimePickerEventBeginTime.Enabled := CheckBoxEventBegin.Checked;
end;
{!******************************************************************************
* MODULE NAME ...  CheckBoxEventEndClick
* SYSTEM NAME ...  AcSELerator
* AUTHOR ........  Kent Myers
*
* FUNCTIONAL DESCRIPTION:  Enable/disable End DateTime controls
*
* INPUT PARAMETERS:
*                    Sender : TObject
*
* OUTPUT PARAMETERS:
*                    None:
* RETURN VALUE:
*                    None:
*
* ERRORS:
*
* APPLICATION NOTES & EXAMPLE:
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2008.09.24  Kent Myers        Original implementation
*******************************************************************************}
procedure TfrmChartTimeSeries.CheckBoxEventEndClick(Sender: TObject);
begin
  DateTimePickerEventEndDate.Enabled := CheckBoxEventEnd.Checked;
  DateTimePickerEventEndTime.Enabled := CheckBoxEventEnd.Checked;
end;
{!******************************************************************************
* MODULE NAME ...  ComboBoxChartTypeChange
* SYSTEM NAME ...  AcSELerator
* AUTHOR ........  Kent Myers
*
* FUNCTIONAL DESCRIPTION:  Revise chart type based on combo box selection
*
* INPUT PARAMETERS:
*                    Sender : TObject
*
* OUTPUT PARAMETERS:
*                    None:
* RETURN VALUE:
*                    None:
*
* ERRORS:
*
* APPLICATION NOTES & EXAMPLE:
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2008.09.24  Kent Myers        Original implementation
*******************************************************************************}

procedure TfrmChartTimeSeries.ComboBoxChartTypeChange(Sender: TObject);
begin
  Assert(FChartTimeSeries.EventData <> nil,'EventData is nil');
  FChartTimeSeries.EventData.InputSettings.ChartStyle := TTimeSeriesChartRenderType(ComboBoxChartType.ItemIndex);
  case FChartTimeSeries.EventData.InputSettings.ChartStyle of
    sctLine:
      begin
        LabelChartTransparency.Enabled := false;
        ScrollBarChartTransparency.Enabled := false;
      end;
    sctArea:
      begin
        LabelChartTransparency.Enabled := true;
        ScrollBarChartTransparency.Enabled := true;
      end;
    else
    begin
      ComboBoxChartType.ItemIndex := 0;      // force a valid type
      ComboBoxChartType.Text := 'Line';
      FChartTimeSeries.EventData.InputSettings.ChartStyle := sctLine;
    end;
  end;
  FChartTimeSeries.SaveChartActiveChannels;  // preserve present active channels
  Display(false);    // must regenerate chart to use new chart style

  FChartTimeSeries.ChartDisplay.Update();
end;
{************************************************************************
* NAME:  ComboBoxDateFormatChange
* FUNCTIONAL DESCRIPTION: Change default import date format option
* INPUT PARAMETERS: not used
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************}
procedure TfrmChartTimeSeries.ComboBoxDateFormatChange(Sender: TObject);
var
  myFormat : string;
begin
  SetDateFormat(ComboBoxDateFormat.Text);
  myFormat := GetDateControlFormat(FDateFormat);
  DateTimePickerEventBeginDate.Format := myFormat;
  DateTimePickerEventEndDate.Format := myFormat;
  ResetDeviceInformation;
  if(FChartType = ctSSI) then
  begin
    FillEventListViewSSI(false);
  end else
  if(FChartType = ctLDP) then
  begin
    FillEventListViewLDP(false);
  end;
  SetDateTimeLabels();
end;

{!******************************************************************************
* MODULE NAME ...  ComboBoxEventTypeExit
* SYSTEM NAME ...  AcSELerator
* AUTHOR ........  Kent Myers
*
* FUNCTIONAL DESCRIPTION:  Test validity of event filter for Event Phase
*
* INPUT PARAMETERS:
*                    Sender : TObject
*
* OUTPUT PARAMETERS:
*                    None:
* RETURN VALUE:
*                    None:
*
* ERRORS:
*
* APPLICATION NOTES & EXAMPLE:
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2008.12.10  Kent Myers        Original implementation
*******************************************************************************}

procedure TfrmChartTimeSeries.ComboBoxEventPhaseExit(Sender: TObject);
var
  phase : integer;
begin
  // phase and event types selected by user
  phase := comboBoxEventPhase.ItemIndex+1;

  if(phase <= 0) then
  begin
    comboBoxEventPhase.ItemIndex := 3;   // set to all
  end;
end;
{!******************************************************************************
* MODULE NAME ...  ComboBoxEventTypeExit
* SYSTEM NAME ...  AcSELerator
* AUTHOR ........  Kent Myers
*
* FUNCTIONAL DESCRIPTION:  Test validity of event filter for Event Type
*
* INPUT PARAMETERS:
*                    Sender : TObject
*
* OUTPUT PARAMETERS:
*                    None:
* RETURN VALUE:
*                    None:
*
* ERRORS:
*
* APPLICATION NOTES & EXAMPLE:
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2008.12.10  Kent Myers        Original implementation
*******************************************************************************}
procedure TfrmChartTimeSeries.ComboBoxEventTypeExit(Sender: TObject);
var
  eventType : integer;
begin
  // phase and event types selected by user
  eventType := comboBoxEventType.ItemIndex+1;

  if(eventType <= 0) then
  begin
    comboBoxEventType.ItemIndex := 3;   // set to all
  end;
end;
{************************************************************************
* NAME:    ComboBoxScaleChange
* FUNCTIONAL DESCRIPTION:  change between primary and secondary units
* INPUT PARAMETERS: not used
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************}
procedure TfrmChartTimeSeries.ComboBoxScaleChange(Sender: TObject);
var
  channelCount : integer;
  channelScale : array of single;
  i,j : integer;
  channelRecord     : TTimeSeriesLDPChannelRecord;
  chartSeries: TChartSeriesList;
  doSecondary : boolean;
begin

  if(ChartType <> ctLDP) then
  begin
    exit;
  end;

  chartSeries := ChartHMIDisplay.SeriesList;
  if(chartSeries.Count < 1) then   // no data loaded into chart
  begin
    exit;
  end;

  if(ComboBoxScale.ItemIndex = -1) then  // invalid selection, reset to Secondary
  begin
    ComboBoxScale.ItemIndex := 1;
    ComboBoxScale.Text := 'Secondary';
  end;

  if(FChartTimeSeries.EventData.InputSettings.ChartSecondaryUnits =
     (ComboBoxScale.Text = 'Secondary')) then   // no change, just exit
  begin
    exit;
  end;

  FChartTimeSeries.EventData.InputSettings.ChartSecondaryUnits := (ComboBoxScale.Text = 'Secondary');

  channelCount := FChartTimeSeries.EventData.LDPChannelTable.Count;  // number of channels to display
  SetLength(ChannelScale,channelCount);    // scale factors for channels
  for j := 0 to FChartTimeSeries.EventData.LDPChannelTable.Count - 1 do
  begin
    channelRecord := (ChartTimeSeries.EventData.LDPChannelTable[j] as TTimeSeriesLDPChannelRecord);  // get event summary record
    channelScale[j] := channelRecord.ChannelScale;
  end;

  doSecondary := ChartTimeSeries.EventData.InputSettings.ChartSecondaryUnits;

  if(doSecondary) then
  begin
    LabelChartToolbar.Caption := '   Units are Secondary';
  end else
  begin
    LabelChartToolbar.Caption := '   Units are Primary';
  end;

  {The chartSeries might contain one more series than the number of channels.
   It might have an event marker series at the end of the list.
  }
  Assert((FChartTimeSeries.ChartVariableCount <= chartSeries.Count),
    'Invalid chart series count.');

  for j := 0 to FChartTimeSeries.ChartVariableCount-1 do        // for each channel
  begin
    for i := 0 to chartSeries[j].Count-1 do        // for point in each channel
    begin
      if(doSecondary) then
      begin
        chartSeries[j].YValue[i] := chartSeries[j].YValue[i]/channelScale[j];
      end else
      begin
        chartSeries[j].YValue[i] := chartSeries[j].YValue[i]*channelScale[j];
      end;
    end;
  end;
  FChartTimeSeries.UpdateEventScale;
  FChartTimeSeries.AutoScaleAxes;
end;

{!******************************************************************************
* MODULE NAME ...  ListViewEventListMouseUp
* SYSTEM NAME ...  AcSELerator
* AUTHOR ........  Kent Myers
*
* FUNCTIONAL DESCRIPTION:  Identify which check box was selected in list view
*
* INPUT PARAMETERS:
*                    Sender : TObject
*                    Button : TMouseButton
*                    Shift  : TShiftState
*                    X,Y    : screen position of cursor
*
* OUTPUT PARAMETERS:
*                    None:
* RETURN VALUE:
*                    None:
*
* ERRORS:
*
* APPLICATION NOTES & EXAMPLE:
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2008.09.24  Kent Myers        Original implementation (obtained from internet)
*******************************************************************************}
procedure TfrmChartTimeSeries.ListViewEventListMouseUp(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  sindex : string;
  indexListView : integer;
  item: TListItem;
  hitTest: THitTests;
  indexRecord : TTimeSeriesIndexRecord;
  more : boolean;
  imin, imax, ipos, indexEventData : integer;
begin
  // Which item belongs to the checkbox
  item := ListViewEventList.GetItemAt(x, y);

  // What kind of thing was hit on the item
  hitTest := ListViewEventList.GetHitTestInfoAt(x, y);

  // If an Item was hit and exactly his checkbox
  if (item <> nil) and (hitTest = [htOnStateIcon]) then
  begin
    sindex := AnsiReplaceStr(Item.Caption,'#',' ');  // list view might have been sorted by some column
    indexListView := StrToInt(sindex)-1;

    // binary search for matching event record

    imin := 0;
    imax := FChartTimeSeries.EventData.IndexTable.Count-1;
    more := true;
    ipos := 0;
    while more do
    begin
      ipos := (imin + imax) div 2;
      indexRecord := (FChartTimeSeries.EventData.IndexTable[ipos] as TTimeSeriesIndexRecord);  // get event summary record
      sindex := AnsiReplaceStr(IndexRecord.EventName,'#',' ');  // list view might have been sorted by some column
      indexEventData := StrToInt(sindex)-1;
      if(indexListView > indexEventData) then     // in top half
      begin
        imin := ipos;
      end else                    // in bottom half
      begin
        imax := ipos;
      end;
      if((imax - imin) < 2) then
      begin
        more := false;
      end;
    end;

    if(imin <> imax) then
    begin
      indexRecord := (FChartTimeSeries.EventData.IndexTable[imin] as TTimeSeriesIndexRecord);  // get event summary record
      sindex := AnsiReplaceStr(IndexRecord.EventName,'#',' ');  // list view might have been sorted by some column
      indexEventData := StrToInt(sindex)-1;
      if(indexListView <> indexEventData) then
      begin
        ipos := imax;
      end else
      begin
        ipos := imin;
      end;
    end;
    indexRecord := (FChartTimeSeries.EventData.IndexTable[ipos] as TTimeSeriesIndexRecord);  // get event summary record
    indexRecord.Checked := item.Checked;
    UpdateChart(true);
  end;
end;

{!******************************************************************************
* MODULE NAME ...  ListViewEventListColumnClick
* SYSTEM NAME ...  AcSELerator
* AUTHOR ........  Kent Myers
*
* FUNCTIONAL DESCRIPTION:  sort by selected column in list view
*
* INPUT PARAMETERS:
*                    Sender : TObject
*                    Column : TListColumn   column to sort by
*
* OUTPUT PARAMETERS:
*                    None:
* RETURN VALUE:
*                    None:
*
* ERRORS:
*
* APPLICATION NOTES & EXAMPLE:
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2008.09.30  Kent Myers        Original implementation
*******************************************************************************}
procedure TfrmChartTimeSeries.ListViewEventListColumnClick(Sender: TObject; Column: TListColumn);
begin
  if Column.Index = FChartTimeSeries.EventLastSortedColumn then
  begin
    FChartTimeSeries.EventAscending := not ChartTimeSeries.EventAscending
  end else
  begin
    FChartTimeSeries.EventLastSortedColumn := Column.Index;
  end;

  TListView(Sender).CustomSort(nil, Column.Index);   // Calls OnCompare()
end;
{************************************************************************
* NAME:  ListViewEventListSelectItem
* FUNCTIONAL DESCRIPTION:  Display selected event in chart
* INPUT PARAMETERS: Sender not used
* OUTPUT PARAMETERS: not used
* RETURN: not used
************************************************************************}
procedure TfrmChartTimeSeries.ListViewEventListSelectItem(Sender: TObject;
  Item: TListItem; Selected: Boolean);
var
  indexRecord : TTimeSeriesIndexRecord;
begin
  if(not Item.Checked) then     // only change state if item is not checked
  begin
    indexRecord := (FChartTimeSeries.EventData.IndexTable[Item.Index] as TTimeSeriesIndexRecord);  // get event summary record
    indexRecord.Checked := Item.Selected;
  end;
  if(Item.Selected) then   // only update chart if item is selected
  begin
    UpdateChart(true);
  end;
end;

{!******************************************************************************
* MODULE NAME ...  ListViewEventListCompare
* SYSTEM NAME ...  AcSELerator
* AUTHOR ........  Kent Myers
*
* FUNCTIONAL DESCRIPTION:  compare function for sorting listview by column
*
* INPUT PARAMETERS:  Sender : TObject   unused
*                    Item1, Item2 : TListItem   items to be compared
*                    Data : integer   data column to be compared in rows Item1,Item2
*
* OUTPUT PARAMETERS:
*                    None:
* RETURN VALUE:
*                    None:
*
* ERRORS:
*
* APPLICATION NOTES & EXAMPLE:
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2008.09.30  Kent Myers        Original implementation
*******************************************************************************}
procedure TfrmChartTimeSeries.ListViewEventListCompare(Sender: TObject; Item1,
  Item2: TListItem; Data: Integer; var Compare: Integer);
const
  COLUMN_ONE = 0;
begin
  if Data = COLUMN_ONE then    // first column uses caption instead of subitem
  begin
    Compare := AnsiCompareText(Item1.Caption, Item2.Caption)
  end else
  begin
    Compare := AnsiCompareText(Item1.SubItems[Data-1],
                              Item2.SubItems[Data-1]);
  end;
  if not FChartTimeSeries.EventAscending then
  begin
    Compare := -Compare;
  end;
end;

{************************************************************************
* NAME:    TabSetChartChange
* FUNCTIONAL DESCRIPTION: select type of chart to display for SSI
* INPUT PARAMETERS: Sender : TObject
*                   NewTab : Integer   selected tab
* OUTPUT PARAMETERS:
*                   AllowChange : Boolean
* RETURN: none
************************************************************************}
procedure TfrmChartTimeSeries.TabSetChartChange(Sender: TObject; NewTab: Integer;
  var AllowChange: Boolean);
const
CBEMA_TAB_INDEX = 0;
var
  reset,DoCBEMA : boolean;
begin
  Assert(FChartTimeSeries.EventData <> nil,'EventData is nil');

  FChartTimeSeries.SaveChartActiveChannels;
  // read input controls
  (FChartTimeSeries.EventData.InputSettings as TTimeSeriesInputSettingsSSI).GetChartInputControlSettings();
  DoCBEMA := (NewTab = CBEMA_TAB_INDEX);
  reset := CBEMAChartVisible(doCBEMA);
  if(DoCBEMA) then
  begin
    ChartHMICBEMA.SetFocus;
  end else
  begin
    ChartHMIDisplay.SetFocus;
  end;
  UpdateChart(reset);
end;
{************************************************************************
* NAME:           GetChartInputControlSettingsSSI
* FUNCTIONAL DESCRIPTION:   getter for Chart input controls
* INPUT PARAMETERS: none
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************}
procedure TfrmChartTimeSeries.GetChartInputControlSettingsSSI();
const
CBEMA_TAB_INDEX = 0;
begin
  ChartInputSettingsSSI.BeginDateActive := CheckBoxEventBegin.Checked;
  ChartInputSettingsSSI.EndDateActive := CheckBoxEventEnd.Checked;
  ChartInputSettingsSSI.ChartCBEMA := (TabSetChart.TabIndex = CBEMA_TAB_INDEX);

  ChartInputSettingsSSI.BeginDateTime := DateOf(dateTimePickerEventBeginDate.Date) +
                   TimeOf(dateTimePickerEventBeginTime.Time);
  ChartInputSettingsSSI.EndDateTime := DateOf(dateTimePickerEventEndDate.Date) +
                 TimeOf(dateTimePickerEventEndTime.Time);

  // phase and event types selected by user
  ChartInputSettingsSSI.EventPhase := THMIEventPhase(comboBoxEventPhase.ItemIndex);
  ChartInputSettingsSSI.EventType := TTimeSeriesEventType(comboBoxEventType.ItemIndex);

end;
{************************************************************************
* NAME:           SetChartInputControlSettingsSSI
* FUNCTIONAL DESCRIPTION:   setter for Chart input controls
* INPUT PARAMETERS: none
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************}
procedure TfrmChartTimeSeries.SetChartInputControlSettingsSSI();
const
CBEMA_TAB_INDEX = 0;
SSI_TAB_INDEX = 1;
begin
  CheckBoxEventBegin.Checked := ChartInputSettingsSSI.BeginDateActive;
  CheckBoxEventEnd.Checked := ChartInputSettingsSSI.EndDateActive;
  if(ChartInputSettingsSSI.ChartCBEMA) then
  begin
    TabSetChart.TabIndex := CBEMA_TAB_INDEX;   // CBEMA
  end else
  begin
    TabSetChart.TabIndex := SSI_TAB_INDEX;   // SSI
  end;

  comboBoxEventPhase.ItemIndex := integer(ChartInputSettingsSSI.EventPhase);
  comboBoxEventType.ItemIndex := integer(ChartInputSettingsSSI.EventType);
end;
{************************************************************************
* NAME:           GetChartInputControlSettings
* FUNCTIONAL DESCRIPTION:   getter for Chart input controls
* INPUT PARAMETERS: none
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************}
procedure TfrmChartTimeSeries.GetChartInputControlSettingsLDP();
begin
  ChartInputSettingsLDP.BeginDateActive := CheckBoxEventBegin.Checked;
  ChartInputSettingsLDP.EndDateActive := CheckBoxEventEnd.Checked;

  ChartInputSettingsLDP.BeginDateTime := DateOf(dateTimePickerEventBeginDate.Date) +
                   TimeOf(dateTimePickerEventBeginTime.Time);
  ChartInputSettingsLDP.EndDateTime := DateOf(dateTimePickerEventEndDate.Date) +
                 TimeOf(dateTimePickerEventEndTime.Time);
end;
{************************************************************************
* NAME:           SetChartInputControlSettings
* FUNCTIONAL DESCRIPTION:   setter for Chart input controls
* INPUT PARAMETERS: none
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************}
procedure TfrmChartTimeSeries.SetChartInputControlSettingsLDP();
begin
  CheckBoxEventBegin.Checked := ChartInputSettingsLDP.BeginDateActive;
  CheckBoxEventEnd.Checked := ChartInputSettingsLDP.EndDateActive;
end;
{************************************************************************
* NAME:     SetDateFormat
* FUNCTIONAL DESCRIPTION:  Set the date format used for parsing dates
* INPUT PARAMETERS: aDateFormat : string   MDY, YMD, DMY
* OUTPUT PARAMETERS: none
* RETURN:           none
************************************************************************}
procedure TfrmChartTimeSeries.SetDateFormat(aDateFormat: string);
begin
  FDateFormat := GetDateFormat(aDateFormat);
  FChartTimeSeries.EventData.DateFormat := FDateFormat;
  ComboBoxDateFormat.Text := aDateFormat;
end;
{************************************************************************
* NAME:     GetDateFormat
* FUNCTIONAL DESCRIPTION:  Get the date format used for parsing dates
* INPUT PARAMETERS: aDateFormat : string   MDY, YMD, DMY
* OUTPUT PARAMETERS: none
* RETURN:           none
************************************************************************}
function TfrmChartTimeSeries.GetDateFormat(aDateFormat: string) : string;
var
  myDateFormat : string;
begin
  if((aDateFormat = '') or (aDateFormat = 'MDY')) then
  begin
    myDateFormat := 'mm/dd/yyyy';
  end else
  if(aDateFormat = 'DMY') then
  begin
    myDateFormat := 'dd/mm/yyyy';
  end else
  if(aDateFormat = 'YMD') then
  begin
    myDateFormat := 'yyyy/mm/dd';
  end else
  begin
    Assert(false,'Invalid date format');
  end;
  result := myDateFormat;
end;
{************************************************************************
* NAME:     UpdateChartHandler
* FUNCTIONAL DESCRIPTION:  Update chart from content of report window
* INPUT PARAMETERS: none
* OUTPUT PARAMETERS: none
* RETURN:           none
************************************************************************}
procedure TfrmChartTimeSeries.UpdateChartHandler();
var
  success : boolean;
begin
  success := DisplayChart(Trim(FNotepadForm.RichEditReport.Text));
  if(success) then
  begin
    Caption := Format('Chart %d - from Notepad',[FChartInstance]);
  end else
  begin
    ShowMessage('Chart generation failed.');
  end;
end;
{************************************************************************
* NAME:     UpdateReportHandler
* FUNCTIONAL DESCRIPTION:  Update report from content of chart window
* INPUT PARAMETERS: none
* OUTPUT PARAMETERS: none
* RETURN:           none
************************************************************************}
procedure TfrmChartTimeSeries.UpdateReportHandler();
var
  buffer : string;
  success : boolean;
begin
  success := AppendReport(buffer);
  if(success) then
  begin
    FNotepadForm.RichEditReport.Text := buffer;
  end else
  begin
    ShowMessage('Report generation failed.');
  end;
end;
{**********************************************************************
* MODULE NAME ...   GetSpecialFolder
* SYSTEM NAME ...   503X
* AUTHOR ........   Nick Frazier
*
* FUNCTIONAL DESCRIPTION:  Makes a call to the Windows API to
* get the path of a special folder.
*
* PARAMETERS: csidl : Integer - A Windows OS defined integer that
* identifies a special Windows folder.
*
* RETURN VALUE: string - the path to the specified folder.
*
* ERRORS: Can throw ESystemFolderException if call to SHFolder.dll
* fails.
*
* APPLICATION EXAMPLE & NOTES:
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/    REVISION
* --DATE--    IMPLEMENTOR    DESCRIPTION
* 2006.09.20  Nick Frazier   Copied and modified from DBConnection.pas
* 2007.02.22  Nick Frazier   Remove unnecessary line of code, use StrPas
*                            on result.
* 2009.07.13  Kent Myers     Copied from SELUtilitiesFunc to avoid a
*                            dependency on SELDevice.
***********************************************************************}
function TfrmChartTimeSeries.GetSpecialFolder(csidl: Integer): String;
var
  sysFolderFunction: TSystemFolderFunction;
  DLLHandle: THandle;
  folderPath: array[0..MAX_PATH] of Char;
begin
  DLLHandle := LoadLibrary('SHFolder.dll');

  if (DLLHandle = 0) then
  begin
    raise ESystemFolderException.Create('Cannot load SHFolder.dll');
  end;

  try
    @sysFolderFunction := GetProcAddress(DLLHandle,PChar('SHGetFolderPathA'));
    if Not (Succeeded(sysFolderFunction(Application.Handle, csidl, 0,
        SHGFP_TYPE_CURRENT, folderPath))) or (StrLen(folderPath) = 0) then
    begin
      raise ESystemFolderException.Create('Unable to locate system folder.');
    end;

    Result := StrPas(folderPath);
  finally
    if DLLHandle <> 0 then
    begin
      FreeLibrary(DLLHandle);
    end;
  end;
end;
end.
